/* Target-vector operations for controlling Mach 3.0 tasks and threads.
   Copyright (c) 1991, 1992, 1993 Free Software Foundation, Inc.

This file is part of GDB.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*
 * HISTORY
 * $Log:	os-mach3.c,v $
 * Revision 2.1.1.1  93/04/16  15:39:16  pds
 * 	Re-worked exception handling.  In particular, removed all usages
 * 	of thread exception ports.  While not providing much functionality,
 * 	they made it difficult to debug applications that used exceptions.
 * 	Added 'task next-exception' command for use when multiple
 * 	exceptions occur.
 * 	Ignore SIGCHLD exceptions while starting an inferior.
 * 	[93/04/09            pds]
 * 
 */

/*
 * Author: Peter Stout <pds@cs.cmu.edu>
 *	   School of Computer Science
 *	   Carnegie Mellon University
 *	   Pittsburgh, PA, USA
 *
 * This work is based on previous versions by:
 *
 * Author: Jukka Virtanen <jtv@hut.fi>
 *	   Computing Centre
 *         Helsinki University of Technology
 *         Finland
 *
 * Thanks to my friends who helped with ideas and testing:
 *
 *	Johannes Helander, Antti Louko, Tero Mononen,
 *	jvh@cs.hut.fi	   alo@hut.fi   tmo@cs.hut.fi
 *
 *      Tero Kivinen       and          Eamonn McManus
 *	kivinen@cs.hut.fi               emcmanus@gr.osf.org
 *	
 */

#include <stdio.h>

#include <mach.h>
#include <mach_error.h>
#include <mach/exception.h>
#include <mach/message.h>
#include <mach/mig_errors.h>
#include <mach/notify.h>
#include <mach/thread_switch.h>
#include <mach/vm_attributes.h>
#include <servers/netname.h>
#include <servers/machid.h>
#include <servers/machid_lib.h>
#include <sys/errno.h>
#include <sys/ptrace.h>
#include <sys/syscall.h>
#include <cthreads.h>
#include <ctype.h>
#include <signal.h>
#include <string.h>

#include "defs.h"
#include "command.h"
#include "gdbcore.h"
#include "gdbcmd.h"
#include "inferior.h"
#include "language.h"
#include "os-mach3.h"
#include "symtab.h"
#include "target.h"
#include "wait.h"

#ifndef	private
#define	private	static
#endif

/*
 * Miscellaneous Mach 3.0 stuff that isn't really exported, but GDB
 * needs to know.
 */

/*
 * The 'approved hack' for getting the privileged host port.  The
 * MachId server uses this port to check authentication.  If GDB can
 * get it, then you can debug any task on the host.  If GDB can't
 * get this port, it uses its task port for authentication.
 */
#define  mach_privileged_host_port() task_by_pid(-1)

/*
 * This is the MIG ID number of the emulator/server bsd_execve() RPC call.
 *
 * It SHOULD never change, but if it does, the GDB `run' command won't
 * work until you fix this define.
 * 
 */
#define MIG_EXEC_SYSCALL_ID		101000

/*
 * This is not defined by GDB's wait.h, but we need it.
 */
#ifndef	WNOHANG
#define	WNOHANG		1
#endif

/*
 * These are the 3.0 cproc states.  The 2.5 state SPINNING maps to
 * switching, which shouldn't cause to much confusion.
 */
#define	CPROC_RUNNING	0
#define CPROC_SWITCHING	1
#define	CPROC_BLOCKED	2
#define	CPROC_CONDWAIT	4

/*
 * This matches the definition for Mach 3.0 with WAIT_DEBUG enabled.
 */
typedef struct cproc_3 {
    struct cproc_3 *	next;
    cthread_t		incarnation;
    struct cproc_3 *	list;
    char *		waiting_for;
    mach_port_t		reply_port;
    int			context;
    spin_lock_t		lock;
    int			state;
    mach_port_t		wired;
    int			busy;
    mach_msg_header_t	msg;
    unsigned int	stack_base;
    unsigned int	stack_size;
} cproc_3_t;

/*
 * This matches the definition for Mach 2.5 MTHREAD.
 * (Assuming that sizeof(port_t) == sizeof(mach_port_t).)
 */
typedef struct cproc_2 {
    struct cproc_2 *	next;
    cthread_t		incarnation;
    int			state;
    mach_port_t		reply_port;
    mach_port_t		wait_port;
    int			id;
    struct cproc_2 *	list;
    int			flags;
    unsigned int	stack_base;
    unsigned int	stack_size;
} cproc_2_t;

/*
 * This matches the definition for Mach 2.5.  The mutex and condition
 * fields have been explicitly declared using the old definitions.
 */
typedef struct cthread_2 {
    struct cthread_2 *	next;
    int			lock_lock;
    char *		lock_name;
    int			done_lock;
    struct cthread_2 *	done_queue_head;
    struct cthread_2 *	done_queue_tail;
    char *		done_name;
    int			state;
    jmp_buf		catch;
    cthread_fn_t	func;
    char *		arg;
    char *		result;
    char *		name;
    char *		data;
} cthread_2_t;

/*
 * All the Mach 3.0 information relating to the target.
 */

typedef struct exc_rec {
    struct exc_rec *	next;
    mach_id_t		mid;
    thread_t		thread;
    int			exception;
    int			code;
    int			sub_code;
} exc_rec_t;

typedef struct gdb_thread_state {
    unsigned int	kernel: 8;
    unsigned int	unused: 6;
    unsigned int	swapped: 1;
    unsigned int	has_cthread: 1;
    unsigned int	cthread: 16;
} gdb_thread_state_t;

typedef struct gdb_thread {
    thread_t		port;
    mach_id_t		mid;
    boolean_t		suspended;
    boolean_t		aborted;
    CORE_ADDR		pc;
    CORE_ADDR		sp;
    CORE_ADDR		fp;
    int			suspend_count;
    char *		name;
    boolean_t		in_emulator;
    CORE_ADDR		user_sp;
    CORE_ADDR		user_pc;
    CORE_ADDR		user_fp;
    gdb_thread_state_t	state;
    boolean_t		wired;
    unsigned int	stack_base;
    unsigned int	stack_top;
} gdb_thread_t;

struct inferior {
    task_t		task;
    mach_id_t		mid;
    boolean_t		emulator_present;
    boolean_t		single_stepping;
    boolean_t		suspended;
    boolean_t		setup;
    boolean_t		has_ptrace_server;
    boolean_t		stopped_in_server;
    boolean_t		in_setup;
    unsigned int	thread_count;
    gdb_thread_t *	thread_list;
    unsigned int	cthread_count;
    gdb_thread_t *	cthread_list;
    mach_id_t		mid_counter;
    exc_rec_t *		exc_list;
    mach_port_t		notify_port;
    mach_port_t		exception_port;
    mach_port_t		old_exception_port;
} inferior;

/*
 * Data type used by GDB to record the ports for which it has
 * requested dead-name notification.
 */

typedef struct port_chain {
  struct port_chain *	next;
  mach_port_t		port;
  mach_type_t		type;
  mach_id_t		mid;
} *port_chain_t;

typedef struct exception_msg {
    mach_msg_header_t head;
    mach_msg_type_t thread_type;
    mach_port_t thread;
    mach_msg_type_t task_type;
    mach_port_t task;
    mach_msg_type_t exception_type;
    int exception;
    mach_msg_type_t code_type;
    int code;
    mach_msg_type_t sub_code_type;
    int sub_code;
} exception_msg_t;

typedef struct exception_reply {
    mach_msg_header_t head;
    mach_msg_type_t ret_code_type;
    int ret_code;
} exception_reply_t;

/*
 * Imported functions.
 */

extern int
grab_terminal PARAMS ((void));		/* XXX - should be elsewhere. */

extern int
getpid PARAMS ((void));

extern void
message ();				/* XXX - should be elsewhere. */

extern int
kill PARAMS ((int, int));

extern mach_msg_return_t
mach_msg_send PARAMS ((mach_msg_header_t *));

extern boolean_t
notify_server PARAMS ((mach_msg_header_t *, mach_msg_header_t *));

extern int
pid_by_task PARAMS ((task_t, int *, char *, int *));

extern int
ptrace PARAMS ((int, int, int *, int));

extern int
syscall ();

extern task_t
task_by_pid PARAMS ((int));

extern kern_return_t
thread_switch PARAMS ((thread_t, int, int));

/*
 * Imported target dependent functions.
 */

extern boolean_t
mach3_fetch_user_state PARAMS ((CORE_ADDR, CORE_ADDR *, CORE_ADDR *,
				CORE_ADDR *));

extern void
mach3_fetch_registers PARAMS ((int));

extern void
mach3_store_registers PARAMS ((int));

/*
 * Functions exported though mach3_ops.
 */

private void
mach3_attach PARAMS ((char *, int));

private int
mach3_can_run PARAMS ((void));

private void
mach3_close PARAMS ((int));

private void
mach3_create_inferior PARAMS ((char *, char *, char **));

private void
mach3_detach PARAMS ((char *, int));

private void
mach3_files_info PARAMS ((struct target_ops *));

private void
mach3_kill PARAMS ((void));

private void
mach3_mourn_inferior PARAMS ((void));

private void
mach3_open PARAMS ((char *, int));

private void
mach3_prepare_to_store PARAMS ((void));

private void
mach3_resume PARAMS ((int, int));

private int
mach3_wait PARAMS ((WAITTYPE *));

private int
mach3_xfer_memory PARAMS ((CORE_ADDR, char *, int, int, struct target_ops *));

/*
 * Functions exported via the command inferface.
 */

private void
exception_command PARAMS ((char *, int));

private void
exception_info PARAMS ((char *, int));

private void
task_command PARAMS ((char *, int));

private void
task_next_exception_command PARAMS ((char *, int));

private void
task_resume_command PARAMS ((char *, int));

private void
task_suspend_command PARAMS ((char *, int));

private void
task_info_command PARAMS ((char *, int));

private void
port_name_set_command PARAMS ((char *, int));

private void
port_name_show_command PARAMS ((char *, int));

private void
thread_command PARAMS ((char *, int));

private void
thread_kill_command PARAMS ((char *, int));

private void
thread_list_command PARAMS ((char *, int));

private void
thread_resume_command PARAMS ((char *, int));

private void
thread_select_command PARAMS ((char *, int));

private void
thread_suspend_command PARAMS ((char *, int));

/*
 * Internal functions.
 */

private kern_return_t
our_ptrace PARAMS ((int, int));

private void
consume_thread_list PARAMS ((thread_t *, mach_msg_type_number_t));

private kern_return_t
gdb_exception_catch PARAMS ((mach_port_t, mach_port_t, mach_msg_type_name_t,
			     task_t, thread_t, int, int, int));
private kern_return_t
gdb_exception_raise PARAMS ((mach_port_t, mach_msg_type_name_t, mach_port_t,
			     mach_msg_type_name_t, thread_t, task_t, int,
			     int, int));

private boolean_t
gdb_exception_server PARAMS ((exception_msg_t *, exception_reply_t *));

private boolean_t
gdb_message_server PARAMS ((mach_msg_header_t *));

private boolean_t
have_emulator_p PARAMS ((task_t));

private void
inferior_check PARAMS ((void));

private gdb_thread_t *
inferior_get_cthreads PARAMS ((unsigned int *));

private void
inferior_init PARAMS ((void));

private void
inferior_exception_discard PARAMS ((void));

private exc_rec_t *
inferior_exception_next PARAMS ((void));

private void
inferior_prepare PARAMS ((int));

private boolean_t
inferior_process_startup_msg PARAMS ((mach_msg_header_t *,mach_msg_header_t *));

private int
inferior_read PARAMS ((CORE_ADDR, char *, int));

private void
inferior_restore PARAMS ((int));

private void
inferior_resume PARAMS ((int));

private void
inferior_setup PARAMS ((void));

private void
inferior_setup_cthreads PARAMS ((void));

private void
inferior_setup_self PARAMS ((void));

private kern_return_t
inferior_suspend PARAMS ((void));

private int
inferior_write PARAMS ((CORE_ADDR, char *, int));

private CORE_ADDR
lookup_address_of_variable PARAMS ((char *));

private void *
mach3_thread PARAMS ((void *));

private mach_id_t
map_port_name_to_mid PARAMS ((mach_port_t, mach_type_t));

private void
notify_cancel PARAMS ((mach_port_t));

private void
notify_request PARAMS ((mach_port_t, mach_type_t));

private void
port_chain_delete PARAMS ((port_chain_t *, mach_port_t));

private void
port_chain_destroy PARAMS ((struct obstack *));

private void
port_chain_insert PARAMS ((port_chain_t *, mach_port_t, mach_type_t));

private port_chain_t
port_chain_lookup PARAMS ((port_chain_t, mach_port_t));

private boolean_t
port_valid PARAMS ((mach_port_t, mach_port_type_t));

private void
print_exception PARAMS ((char *, int, int, int));

private void
print_pc PARAMS ((FILE *stream, CORE_ADDR));

private void
process_msg PARAMS ((mach_msg_header_t *, mach_msg_header_t *));

private void
sigchld_handler PARAMS ((int));

private void
thread_cleanup PARAMS ((thread_t));

private void
thread_init PARAMS ((thread_t));

private gdb_thread_t *
thread_lookup PARAMS ((thread_t));

private gdb_thread_t *
thread_lookup_by_mid PARAMS ((mach_id_t));

private void
thread_restore PARAMS ((gdb_thread_t *));

private void
thread_resume_all PARAMS ((void));

private kern_return_t
thread_setup PARAMS ((mach_id_t));

private char *
thread_state_to_string PARAMS ((gdb_thread_t *));

private void
thread_suspend_all PARAMS ((void));

private void
thread_trace PARAMS ((mach_port_t, boolean_t));

private void
wakeup_main_thread PARAMS ((void));

/*
 * Misc.
 */

private char *
get_size PARAMS ((int));

private void
print_exception_info PARAMS ((int));

private void
add_mach_specific_commands PARAMS ((void));

/*
 * Global Variables.
 */

private struct target_ops mach3_ops = {
  "mach3",			/* to_shortname */
  "Mach 3 child process",	/* to_longname */
  "Mach 3 child process (started by the \"run\" command).",	/* to_doc */
  mach3_open,			/* to_open */
  mach3_close,			/* to_close */
  mach3_attach,			/* to_attach */
  mach3_detach, 		/* to_detach */
  mach3_resume,			/* to_resume */
  mach3_wait,			/* to_wait */
  mach3_fetch_registers,	/* to_fetch_registers */
  mach3_store_registers,	/* to_store_registers */
  mach3_prepare_to_store,	/* to_prepare_to_store */
  mach3_xfer_memory,		/* to_xfer_memory */
  mach3_files_info,		/* to_files_info */
  memory_insert_breakpoint,	/* to_insert_breakpoint */
  memory_remove_breakpoint,	/* to_remove_breakpoint */
  terminal_init_inferior,	/* to_terminal_init */
  terminal_inferior, 		/* to_terminal_inferior */
  terminal_ours_for_output,	/* to_terminal_ours_for_output */
  terminal_ours,		/* to_terminal_ours */
  child_terminal_info,		/* to_terminal_info */
  mach3_kill,			/* to_kill */
  0,				/* to_load */
  0,				/* to_lookup_symbol */
  mach3_create_inferior,	/* to_create_inferior */
  mach3_mourn_inferior,		/* to_mourn_inferior */
  mach3_can_run,		/* to_can_run */
  0, 				/* to_notice_signals */
  process_stratum,		/* to_stratum */
  0,				/* to_next */
  TRUE,				/* to_has_all_memory */
  TRUE,				/* to_has_memory */
  TRUE,				/* to_has_stack */
  TRUE,				/* to_has_registers */
  TRUE,				/* to_has_execution */
  0,				/* sections */
  0,				/* sections_end */
  OPS_MAGIC			/* to_magic */
};

/*
 * Task and thread command lists.
 */
private struct cmd_list_element *cmd_task_list;
private struct cmd_list_element *cmd_thread_list;

/*
 * When creating the inferior fork, inferior_setup_self() saves the name
 * of the inferior's bootstrap_port in this variable.
 *
 * The name is transferred to our address space with inferior_read().
 *
 * Thou shalt not do this with task_get_bootstrap_port() in this task,
 * since the name in the inferior task is different than the one we get.
 *
 * NOTE: We could probably "guess" the name of the bootstrap, but we'll
 * be good ...
 */
private mach_port_t gdb_ux_server_port_name = MACH_PORT_NULL;

/*
 * True if GDB is connected to a U*x server that has ptrace support.
 */
private boolean_t	gdb_has_ptrace_server = TRUE;

private struct mutex	gdb_quit_lock = MUTEX_INITIALIZER;
private boolean_t	gdb_really_waiting;
private cthread_t	gdb_main_thread_id;

private mach_port_t	gdb_port_set;

#define	MAX_NAME_LEN	64
private boolean_t	gdb_name_registered;
private char		gdb_port_name[ MAX_NAME_LEN ] = GDB_NAME_PREFIX;
private mach_port_t	gdb_message_port;
private mach_port_t	mid_server;
private mach_port_t	mid_auth;

/*
 * Needed by various machine dependent routines.
 */
thread_t		current_thread;
/*
 * This is filled in by the initialization routine in *-nat.c if we
 * are built in a NATIVE configuration.
 */
void (*mach3_fork_inferior_func) PARAMS ((char *, char *, char **,
					  void (*)(void), void (*)(int)));
/* 
 * Actions for Mach exceptions.
 *
 * sigmap field maps the exception to corresponding Unix signal.
 *
 * I do not know how to map the exception to unix signal
 * if SIG_UNKNOWN is specified.
 */

#define SIG_UNKNOWN 0	/* Exception that has no matching unix signal. */

private struct exception_list {
  char *name;
  boolean_t forward;
  boolean_t print;
  int       sigmap;
} exception_map[] = {
  {"not_mach3_exception",	FALSE, TRUE,  SIG_UNKNOWN},
  {"EXC_BAD_ACCESS",		FALSE, TRUE,  SIGSEGV},
  {"EXC_BAD_INSTRUCTION",	FALSE, TRUE,  SIGILL},
  {"EXC_ARITHMETIC",		FALSE, TRUE,  SIGFPE},
  {"EXC_EMULATION",		FALSE, TRUE,  SIGEMT},	/* ??? */
  {"EXC_SOFTWARE",		FALSE, TRUE,  SIG_UNKNOWN},
  {"EXC_BREAKPOINT",		FALSE, FALSE, SIGTRAP}
};

/*
 * Exception codes used to translate U*x wait state into Mach exceptions.
 */
#define	EXC_GDB_STOPPED		0x20000
#define	EXC_GDB_EXITED		0x20001
#define	EXC_GDB_SIGNALED	0x20002

#define MAX_EXCEPTION	(sizeof(exception_map)/sizeof(exception_map[0]) - 1)

/*
 * Structure and obstack used by write_inferior.
 */
typedef struct vm_region_list {
  struct vm_region_list *next;
  vm_prot_t	protection;
  vm_address_t  start;
  vm_size_t	length;
} vm_region_list_t;

struct obstack  region_obstack;

/*
 * Room for nodes in the port chain comes from pchain_obstack.
 */

struct obstack pchain_obstack;
struct obstack *port_chain_obstack = &pchain_obstack;

port_chain_t notify_chain = (port_chain_t) NULL;

/*
 * Format of the thread_list command
 *
 * 	            slot mid sel   name  emul ks susp  cstate wired   address
 */
#define TL_K_FORMAT "%-4u %5u%c %-10.10s %c%s %-5.5s "
#define TL_C_FORMAT "-%-3u   --   %-10.10s  %s       "

#define TL_HEADER "\n@      MID  Name        KState CState   Where\n"

/*
 * Prefixes used in naming otherwise un-named kernel and C threads.
 */
#define	K_THREAD_PREFIX	"_K"
#define	C_THREAD_PREFIX	"_C"

/*
 * If your char has more than 8-bits change this.
 */
#define	DIGITS_PER_CHAR	3

/*
 * Mach 3.0 kernel does not support ptrace, so make sure nobody uses it.
 */
int
ptrace (a,b,c,d)
int a,b,*c,d;
{
  error ("Lose, Lose! Somebody called ptrace\n");
}

/*
 * However, many people use servers (UX, BSDSS, Multi) that do support
 * ptrace and these servers need to be kept informed about tasks being
 * debugged.
 */
private kern_return_t
our_ptrace (request, send_signal)
     int	request;
     int	send_signal;
{
  int		rc;
  kern_return_t	kr;

  if (inferior.has_ptrace_server)
    {
      rc = syscall (SYS_ptrace, request, inferior_pid, (int *) 1, send_signal);
      if (rc == -1)
        {
	  kr = unix_err(errno);
	  if (kr == unix_err(EINVAL))
	    {
	      gdb_has_ptrace_server = FALSE;
	      inferior.has_ptrace_server = FALSE;
	    }
	  else if (kr == unix_err(EIO) || kr == unix_err(ESRCH))
	    inferior.has_ptrace_server = FALSE;
	}
      else
	kr = KERN_SUCCESS;
    }
  else
    kr = KERN_FAILURE;

  return kr;
}

/*
 * Debugging Tools.
 */

/*
 * If you define DUMP_SYSCALL, intercepted bsd server calls will be
 * dumped while waiting for the inferior to EXEC the correct program.
 */
/* #define DUMP_SYSCALL */

/*
 * xx_debug() prints messages if debug-level is non-zero.
 * If it is greater than 1, DUMP_SYSCALL will dump message contents.
 */
int debug_level = 0;

/* "Temporary" debug stuff */
private void
xx_debug (fmt, a,b,c)
char *fmt;
int a,b,c;
{
  if (debug_level)
    message (fmt, a, b, c);
}

#ifdef DUMP_SYSCALL
#ifdef __STDC__
#define STR(x) #x
#else
#define STR(x) "x"
#endif

private char *bsd1_names[] = {
  "execve", "fork", "take_signal", "sigreturn", "getrusage",
  "chdir", "chroot", "open", "creat", "mknod", "link",
  "symlink", "unlink", "access", "stat", "readlink",
  "chmod", "chown", "utimes", "truncate", "rename",
  "mkdir", "rmdir", "xutimes", "mount", "umount",
  "acct", "setquota", "write_short", "write_long", "send_short",
  "send_long", "sendto_short", "sendto_long", "select", "task_by_pid",
  "recvfrom_short", "recvfrom_long", "setgroups", "setrlimit", "sigvec",
  "sigstack", "settimeofday", "adjtime", "setitimer", "sethostname",
  "bind", "accept", "connect", "setsockopt", "getsockopt",
  "getsockname", "getpeername", "init_process", "table_set", "table_get",
  "pioctl", "emulator_error", "readwrite", "share_wakeup", NULL,
  "maprw_request_it", "maprw_release_it", "maprw_remap", "pid_by_task",
};

int	bsd1_nnames = sizeof(bsd1_names)/sizeof(bsd1_names[0]);

private char *
id_str(id,buf)
    int id;
    char *buf;
{
  char	*p;

  id -= 101000;
  if (id >= 0 && id < bsd1_nnames && (p = bsd1_names[id]) != NULL)
      return p;
  if (id == 1000 /* 102000 */)
    return "psignal_retry";
  if (id == -1000 /* 100000 */)
    return "syscall";
  sprintf(buf,"%d", id+101000);

  return buf;
}

private char *
name_str(name,buf)
    int name;
    char *buf;
{
  switch (name) {
  case MACH_MSG_TYPE_BOOLEAN:
    return "boolean";
  case MACH_MSG_TYPE_INTEGER_16:
    return "short";
  case MACH_MSG_TYPE_INTEGER_32:
    return "long";
  case MACH_MSG_TYPE_CHAR:
    return "char";
  case MACH_MSG_TYPE_BYTE:
    return "byte";
  case MACH_MSG_TYPE_REAL:
    return "real";
  case MACH_MSG_TYPE_STRING:
    return "string";
  default:
    sprintf(buf,"%d",name);
    return buf;
  }
}

private void
print_msg(mp)
mach_msg_header_t	*mp;
{
  char	*fmt_x = "%20s : 0x%08x\n";
  char	*fmt_d = "%20s : %10d\n";
  char	*fmt_s = "%20s : %s\n";
  char	buf[100];

  puts_filtered ("\n");
#define pr(fmt,h,x) printf_filtered(fmt,STR(x),(h).x)
  pr(fmt_x,(*mp),msgh_bits);
  pr(fmt_d,(*mp),msgh_size);
  pr(fmt_x,(*mp),msgh_remote_port);
  pr(fmt_x,(*mp),msgh_local_port);
  pr(fmt_d,(*mp),msgh_kind);
  printf_filtered(fmt_s,STR(msgh_id),id_str(mp->msgh_id,buf));
  
  if (debug_level > 1)
  {
    char	*p,*ep,*dp;
    int		plen;
    p = (char*)mp;
    ep = p+mp->msgh_size;
    p += sizeof(*mp);
    for(; p < ep; p += plen) {
      mach_msg_type_t	*tp;
      mach_msg_type_long_t	*tlp;
      int	name,size,number;
      tp = (mach_msg_type_t*)p;
      if (tp->msgt_longform) {
	tlp = (mach_msg_type_long_t*)tp;
	name = tlp->msgtl_name;
	size = tlp->msgtl_size;
	number = tlp->msgtl_number;
	plen = sizeof(*tlp);
      } else {
	name = tp->msgt_name;
	size = tp->msgt_size;
	number = tp->msgt_number;
	plen = sizeof(*tp);
      }
      printf_filtered("name=%-16s size=%2d number=%7d inline=%d long=%d deal=%d\n",
		      name_str(name,buf),size,number,tp->msgt_inline,
		      tp->msgt_longform, tp->msgt_deallocate);
      dp = p+plen;
      if (tp->msgt_inline) {
	int	l;
	l = size*number/8;
	l = (l+sizeof(long)-1)&~((sizeof(long))-1);
	plen += l;
	print_data(dp,size,number);
      } else {
	plen += sizeof(int*);
      }
      printf_filtered("plen=%d\n",plen);
    }
  }
}

private void
print_data(p,size,number)
    char *p;
    int size;
    int number;
{
  int	*ip;
  short	*sp;
  int	i;

  switch (size) {
  case 8:
    for(i = 0; i < number; i++) {
      printf_filtered(" %02x",p[i]);
    }
    break;
  case 16:
    sp = (short*)p;
    for(i = 0; i < number; i++) {
      printf_filtered(" %04x",sp[i]);
    }
    break;
  case 32:
    ip = (int*)p;
    for(i = 0; i < number; i++) {
      printf_filtered(" %08x",ip[i]);
    }
    break;
  }
  puts_filtered("\n");
}
#endif  DUMP_SYSCALL


/*
 * Implementation utilities, some are exported via MACRO hooks.
 */

/*
 * Deallocate a thread list and the memory it came in.
 */
private void
consume_thread_list (thread_list, thread_count)
     thread_t *			thread_list;
     mach_msg_type_number_t	thread_count;
{
  kern_return_t	kr;
  mach_msg_type_number_t		i;

  if (thread_count <= 0)
    return;

  for (i = 0; i < thread_count; i++)
    if (MACH_PORT_VALID (thread_list[i]))
      {
	kr = mach_port_deallocate (mach_task_self (), thread_list[i]);
	if (kr != KERN_SUCCESS)
	  message ("Error deallocating thread %x: %s.", thread_list[i],
		   mach_error_string (kr));
      }

  kr = vm_deallocate (mach_task_self (), (vm_address_t) thread_list,
		      (vm_size_t) thread_count * sizeof(thread_t));
  if (kr != KERN_SUCCESS)
    message ("Error deallocating thread_list: %s.", mach_error_string (kr));
}

/*
 * This routine is called from gdb_exception_server.
 *
 *
 * The code will pass the exception to the inferior if:
 *
 *   - The task that signaled is not the inferior task.
 *     (e.g. when debugging another debugger)
 *     (??pds?? Given that GDB set's its inferior's exception port,
 *     I'm not sure this is a good example.)
 *
 *   - The user has explicitly requested that the exception be forwarded.
 *     (e.g to the default unix exception handler, which maps
 *     exceptions to signals, or the user has her own exception handler)
 *
 *   - If the thread that signaled is being single-stepped and it
 *     has set it's own exception port and the exception is not
 *     EXC_BREAKPOINT. (Maybe this is not desirable?)
 */

private kern_return_t
gdb_exception_catch (port, reply_port, reply_port_type, task, thread,
		     exception, code, subcode)
     mach_port_t	port, reply_port;
     mach_msg_type_name_t reply_port_type;
     task_t		task;
     thread_t		thread;
     int		exception, code, subcode;
{
  kern_return_t		kr;
  gdb_thread_t *	gdb_thread;
  exc_rec_t *		exc_rec;
  exc_rec_t **		prev_next;

  if (! MACH_PORT_VALID (task))
    {
      /*
       * If the exception was sent and task dies before we receive it,
       * TASK will be MACH_PORT_DEAD.  If this wasn't an internally
       * generated message then we will print a message and ignore
       * the exception.
       */
      if (exception == EXC_SOFTWARE
	  && (code == EXC_GDB_SIGNALED || code == EXC_GDB_EXITED))
	{
	  if (MACH_PORT_VALID (inferior.task))
	    {
	      kr = mach_port_mod_refs (mach_task_self (), inferior.task,
				       MACH_PORT_RIGHT_DEAD_NAME, 1);
	      if (kr != KERN_SUCCESS)
	        message ("Unable to increase ref count for dead task port: %s.",
			 mach_error_string (kr));
	    }
	  task = inferior.task;
	}
      else {
	print_exception ("Unknown task (dead)", exception, code, subcode);
	if (MACH_PORT_VALID (thread))
	  (void) mach_port_deallocate (mach_task_self (), thread);

	return KERN_SUCCESS;
      }
    }

  /*
   * If it is not our inferior, pass it on and return.
   */
  if (task != inferior.task)
    {
      if (! MACH_PORT_VALID (thread))
	thread = MACH_PORT_NULL;

      if (MACH_PORT_VALID (inferior.old_exception_port))
        {
	  kr = gdb_exception_raise (inferior.old_exception_port,
				    MACH_MSG_TYPE_COPY_SEND, reply_port,
				    reply_port_type, task, thread,
				    exception, code, subcode);
	  if (kr != KERN_SUCCESS)
	    message ("Error forwarding exception for unknown task: %s.",
		     mach_error_string (kr));
	}

      if (MACH_PORT_VALID (thread))
	(void) mach_port_deallocate (mach_task_self (), thread);
      if (MACH_PORT_VALID (task))
	(void) mach_port_deallocate (mach_task_self (), task);

      /*
       * Don't send a reply, the new recepient will reply.
       */
      return MIG_NO_REPLY;
    }

  /*
   * At this point we only need the name of the task port; deallocating it
   * here simplifies error handling later.
   *
   * We cannot deallocate the thread port, because the thread may not be
   * recorded in the thread list for the inferior.  We will save the thread
   * port (and reference) in the exception record.  When we go to handle the
   * exception, we will try again to map the thread port into a gdb_thread.
   */
  if (MACH_PORT_VALID (task))
    {
      kr = mach_port_deallocate (mach_task_self (), task);
      if (kr != KERN_SUCCESS)
	message ("Error deallocating task port %x during exception: %s.", task,
		 mach_error_string (kr));
    }

  /*
   * Okay, we know have an exception for our task.  Do we know about
   * the thread that raised the exception?
   */
  gdb_thread = NULL;
  if (MACH_PORT_VALID (thread))
    gdb_thread = thread_lookup (thread);

  if (exception < 0 || exception > MAX_EXCEPTION)
    {
      if (gdb_thread)
        message ("Unknown exception %d received for thread %u.", exception,
		 gdb_thread->mid);
      else
	message ("Unknown exception %d received for task %u.", exception,
		 inferior.mid);
      exception = 0;
    }

  /*
   * Forward the exception to the inferior's original handler if requested.
   *
   * Note: If you have forwarded EXC_BREAKPOINT I trust you know why.
   */
  if (exception_map[exception].forward)
    {
      print_exception (gdb_thread? "Thread" : "Task", exception, code, subcode);

      /*
       * If we come here, we don't want to trace any more, since we
       * will never stop for tracing anyway.
       */
      thread_trace (thread, FALSE);

      if (MACH_PORT_VALID (inferior.old_exception_port))
	{
	  kr = gdb_exception_raise (inferior.old_exception_port,
				    MACH_MSG_TYPE_COPY_SEND, reply_port,
				    reply_port_type, task, thread,
				    exception, code, subcode);
	  if (kr != KERN_SUCCESS)
	    if (gdb_thread)
	      message ("Unable to forward exception for Thread %u: %s.",
		       gdb_thread->mid, mach_error_string (kr));
	    else
	      message ("Unable to forward exception for Task %u: %s.",
		       inferior.mid, mach_error_string (kr));
	  kr = MIG_NO_REPLY;
	}
      else
	kr = KERN_SUCCESS;

      if (MACH_PORT_VALID (thread))
	{
	  kern_return_t	kr2;

	  kr2 = mach_port_deallocate (mach_task_self (), thread);
	  if (kr2 != KERN_SUCCESS)
	    message ("Error deallocating thread port %x during exception: %s.",
		     thread, mach_error_string (kr2));
	}

      return kr;
    }
  
  /*
   * GDB is handling this exception.
   */
  kr = inferior_suspend ();
  if (kr != KERN_SUCCESS)
    message ("Found inferior dead while processing an exception.");

  /*
   * Don't print our internally generated exceptions.
   */
  if (exception_map[exception].print
      && (exception != EXC_SOFTWARE
	  || (code != EXC_GDB_SIGNALED && code != EXC_GDB_EXITED
	      && code != EXC_GDB_STOPPED)))
      print_exception (gdb_thread? "Thread" : "Task", exception, code, subcode);

  /*
   * Save the exception at the end of the exception list.
   */
  exc_rec = (exc_rec_t *) xmalloc (sizeof(exc_rec_t));
  exc_rec->next = NULL;
  exc_rec->thread = thread;
  exc_rec->mid = gdb_thread ? gdb_thread->mid : inferior.mid;
  exc_rec->exception = exception;
  exc_rec->code = code;
  exc_rec->sub_code = subcode;
  prev_next = &inferior.exc_list;
  while (*prev_next)
    prev_next = &(*prev_next)->next;
  *prev_next = exc_rec;
  
  /*
   * If we receive an exception that is not a breakpoint exception,
   * we interrupt the single step.  The trace condition is cleared.
   */
  if (inferior.single_stepping)
    {
      if (thread != current_thread)
	message ("Exception from %s %u while single-stepping thread %u.",
		 (gdb_thread ? "thread" : "task"), exc_rec->mid,
		 map_port_name_to_mid (current_thread, MACH_TYPE_THREAD));
      else if (exception != EXC_BREAKPOINT)
	message ("Single step interrupted by exception.");

      thread_trace (current_thread, FALSE);
    }
  
  return KERN_SUCCESS;
}

/*
 * A non-blocking version of exception_raise.  We don't want to
 * dead-lock with ourselves or wait on slow recepients.
 */
private kern_return_t
gdb_exception_raise (req_port, req_port_type, reply_port, reply_port_type,
		     task, thread, exception, code, sub_code)
     mach_port_t 	req_port;
     mach_msg_type_name_t req_port_type;
     mach_port_t 	reply_port;
     mach_msg_type_name_t reply_port_type;
     task_t		task;
     thread_t		thread;
     int		exception, code, sub_code;
{
  static mach_msg_type_t port_msg_type = { MACH_MSG_TYPE_COPY_SEND, 32, 1,
					   TRUE, FALSE, FALSE, 0 };
  static mach_msg_type_t int_msg_type = { MACH_MSG_TYPE_INTEGER_32, 32, 1,
					  TRUE, FALSE, FALSE, 0 };
  exception_msg_t msg;
  kern_return_t kr;

  msg.thread_type = msg.task_type = port_msg_type;
  msg.thread = thread;
  msg.task = task;

  msg.exception_type = msg.code_type = msg.sub_code_type = int_msg_type;
  msg.exception = exception;
  msg.code = code;
  msg.sub_code = sub_code;

  msg.head.msgh_bits = MACH_MSGH_BITS_COMPLEX|MACH_MSGH_BITS(req_port_type,
							     reply_port_type);
  msg.head.msgh_size = sizeof(msg);
  msg.head.msgh_remote_port = req_port;
  msg.head.msgh_local_port = reply_port;
  msg.head.msgh_seqno = 0;
  msg.head.msgh_id = 2400;

  kr = mach_msg (&msg.head, MACH_SEND_MSG|MACH_SEND_TIMEOUT, sizeof(msg), 0,
		 MACH_PORT_NULL, 0, MACH_PORT_NULL);
  return kr;
}

/*
 * A customized version of exc_server.  By passing the reply port to
 * gdb_exception_catch, GDB is able to cut itself out of the loop when
 * forwarding exception messages.
 */
private boolean_t
gdb_exception_server (in_msg, out_msg)
     exception_msg_t *in_msg;
     exception_reply_t *out_msg;
{
  static mach_msg_type_t port_check = { MACH_MSG_TYPE_PORT_SEND, 32, 1,
					TRUE, FALSE, FALSE, 0 };
  static mach_msg_type_t int_check = { MACH_MSG_TYPE_INTEGER_32, 32, 1,
				       TRUE, FALSE, FALSE, 0 };
  mach_msg_type_name_t reply_port_type;

  reply_port_type = MACH_MSGH_BITS_REMOTE(in_msg->head.msgh_bits);
  out_msg->head.msgh_bits = MACH_MSGH_BITS(reply_port_type, 0);
  out_msg->head.msgh_size = sizeof *out_msg;
  out_msg->head.msgh_remote_port = in_msg->head.msgh_remote_port;
  out_msg->head.msgh_local_port = MACH_PORT_NULL;
  out_msg->head.msgh_seqno = 0;
  out_msg->head.msgh_id = in_msg->head.msgh_id + 100;

  out_msg->ret_code_type = int_check;

  if (in_msg->head.msgh_id != 2400)
    {
      out_msg->ret_code = MIG_BAD_ID;
      return FALSE;
  }

  if ((in_msg->head.msgh_size != sizeof(exception_msg_t))
      || ! (in_msg->head.msgh_bits & MACH_MSGH_BITS_COMPLEX)
      || (* (int *) &in_msg->thread_type != * (int *) &port_check)
      || (* (int *) &in_msg->task_type != * (int *) &port_check)
      || (* (int *) &in_msg->exception_type != * (int *) &int_check)
      || (* (int *) &in_msg->code_type != * (int *) &int_check)
      || (* (int *) &in_msg->sub_code_type != * (int *) &int_check))
    {
      out_msg->ret_code = MIG_BAD_ARGUMENTS;
      return TRUE;
    }

  out_msg->ret_code = gdb_exception_catch (in_msg->head.msgh_local_port,
					   in_msg->head.msgh_remote_port,
					   reply_port_type,
					   in_msg->task, in_msg->thread,
					   in_msg->exception, in_msg->code,
					   in_msg->sub_code);
  return TRUE;
}

/*
 * Handle an internally or externally generated STOP message.
 */
private boolean_t
gdb_message_server (in_msg)
     mach_msg_header_t *in_msg;
{
  if (in_msg->msgh_local_port == gdb_message_port)
    {
      switch (in_msg->msgh_id) {
	case GDB_MESSAGE_ID_STOP:
	  /*
	   * This will get us out of the wait loop, eventually.
	   */
	  if (inferior.in_setup)
	    {
	      mach3_kill ();
	      error ("Received quit message, killed inferior.");
	    }
	  quit_flag = TRUE;
	  break;

	default:
	  message ("Invalid message id %d received, ignored.", in_msg->msgh_id);
	  break;
      }

      return TRUE;
    }

  /* Message not handled by this server */
  return FALSE;
}

/*
 * Check if the emulator exists at task's address space.
 */
private boolean_t
have_emulator_p (task)
     task_t task;
{
  kern_return_t			kr;
  int				vector_start;
  mach_msg_type_number_t	i;
  mach_msg_type_number_t	n;
  boolean_t			has_emulator = FALSE;
#ifdef	MK67
  /*
   * The get_emulation_vector() interface was changed after MK67.
   */
#define	EMUL_VECTOR_COUNT	400	/* Value does not matter too much */
  vm_offset_t	emulation_vector[ EMUL_VECTOR_COUNT ];

  n = EMUL_VECTOR_COUNT;
  kr = task_get_emulation_vector (task, &vector_start, emulation_vector, &n);
#else
  vm_offset_t	*emulation_vector;

  kr = task_get_emulation_vector (task, &vector_start, &emulation_vector, &n);
#endif
  
  if (kr != KERN_SUCCESS)
    {
      mach3_mourn_inferior ();
      error ("Task died will checking for emulator.");
    }

  xx_debug ("%d vectors from %d at 0x%x\n", n, vector_start, emulation_vector);
  
  for(i = 0; i < n; i++)
    {
      vm_offset_t entry = emulation_vector [i];

      /*
       * ??pds?? -- It would be a good idea to make a vm_region() call
       * to check that there is memory where the emulator should be.
       */
      if (EMULATOR_BASE <= entry && entry <= EMULATOR_END)
        {
	  has_emulator = TRUE;
	  break;
	}
      else if (entry)
	{
	  static task_t informed;

	  if (task != informed)
	    {
	      message("Emulation vector address 0x%x outside emulator space",
		      entry);
	      message("Not a UX Emulation vector, proceed at your own risk...");
	      informed = task;
	    }
	}
    }

#ifdef	MK67
#else	/* not MK67 */
  if (n > 0)
    {
      kr = vm_deallocate (mach_task_self (), (vm_address_t) emulation_vector,
			  n * sizeof(vm_offset_t));
      if (kr != KERN_SUCCESS)
	message ("Error deallocating emulation vector: %s.",
		 mach_error_string (kr));
    }
#endif

  return has_emulator;
}

/*
 * Check that the inferior task is still alive.  If not, cleanup
 * and go home.
 */
private void
inferior_check ()
{
  if (! MACH_PORT_VALID (inferior.task))
    error ("Inferior task does not exist.");
  else if (! port_valid (inferior.task, MACH_PORT_TYPE_SEND))
    {
      mach3_mourn_inferior ();
      error ("Inferior died (task port invalid).");
    }
}
    
/*
 * Read the C-Thread data structures from the inferior.
 *
 * FIXME:  This implementation will only work for native debugging and
 * cross-debugging where the sizes of the C-Thread data structures and
 * the byte ordering are the same on host and target.  This is really
 * target dependent code.
 */
private gdb_thread_t *
inferior_get_cthreads (thread_count_ptr)
     unsigned int *	thread_count_ptr;
{
/* Allocation increment for thread_list. */
#define	thread_list_inc	8
  gdb_thread_t *	thread_list;
  unsigned int		thread_list_max;
  unsigned int		thread_count;
  struct cleanup *	old_cleanup;
  CORE_ADDR		cproc_list;
  CORE_ADDR		sym_addr;
  boolean_t		new_threads;
  union {
      cproc_3_t		new;
      cproc_2_t		old;
  } cproc;
  union {
      struct cthread	new;
      cthread_2_t	old;
  } cthread;
  unsigned int		cproc_size;
  unsigned int		cthread_size;
  char			name[MAX_NAME_LEN];
  
  sym_addr = lookup_address_of_variable ("cproc_list");

  if (sym_addr)
    new_threads = TRUE;
  else
    {
      /*
       * cproc_list is not in a file compiled with debugging symbols,
       * but don't give up yet.
       */
      
      sym_addr = lookup_address_of_variable ("cprocs");

      if (sym_addr)
	new_threads = FALSE;
    }

  /*
   * Stripped or no -lthreads loaded or "cproc_list" and "cprocs"  are in
   * the wrong segment.
   */
  if (! sym_addr)
    return NULL;

  /* Get the address of the first cproc in the task */
  if (! inferior_read (sym_addr, (char *) &cproc_list, sizeof(cproc_list)))
    {
      message ("Unable to read %s at 0x%x from inferior.",
	       new_threads ? "cproc_list" : "cprocs", sym_addr);
      return NULL;
    }

  if (cproc_list == 0)
    return NULL;

  /*
   * The caller has given a guess for the number of cthreads and
   * will add room for a few more.
   */
  thread_list_max = *thread_count_ptr + thread_list_inc;
  thread_list = (gdb_thread_t *)xmalloc(thread_list_max * sizeof(gdb_thread_t));
  memset (thread_list, 0, thread_list_max * sizeof(gdb_thread_t));

  old_cleanup = make_cleanup (free_current_contents, &thread_list);

  /*
   * Copy out the list of CPROCS/CTHREADS saving what GDB needs.
   * The CPROCS are chained by the LIST field, not the NEXT field.
   */
  if (new_threads)
    {
      cproc_size = sizeof(cproc.new);
      cthread_size = sizeof(cthread.new);
    }
  else
    {
      cproc_size = sizeof(cproc.old);
      cthread_size = sizeof(cthread.old);
    }

  thread_count = 0;
  while (cproc_list)
    {
      gdb_thread_t *	gdb_thread;

      if (thread_count == thread_list_max)
        {
	  thread_list_max += thread_list_inc;
	  thread_list = (gdb_thread_t *) xrealloc (thread_list,
						   thread_list_max *
						   sizeof(gdb_thread_t));
	  memset (&thread_list[thread_count], 0,
		  thread_list_inc * sizeof(gdb_thread_t));
	}

      if (! inferior_read ((CORE_ADDR) cproc_list, (char *) &cproc, cproc_size))
	{
	  message ("Unable to read C-Thread, returning first %u.",
		   thread_count);
	  if (thread_count == 0)
	    {
	      free (thread_list);
	      thread_list = NULL;
	    }
	  break;
	}
      gdb_thread = &thread_list[thread_count];
      gdb_thread->state.has_cthread = TRUE;
      if (new_threads)
        {
	  gdb_thread->state.cthread = cproc.new.state;
	  gdb_thread->wired = cproc.new.wired;
	  gdb_thread->stack_base = cproc.new.stack_base;
	  gdb_thread->stack_top = cproc.new.stack_base + cproc.new.stack_size;

	  if (cproc.new.context)
	    if (inferior_read ((CORE_ADDR) (cproc.new.context
					    + MACHINE_CPROC_FP_OFFSET),
			       (char *) &gdb_thread->fp, sizeof(gdb_thread->fp))
		&& inferior_read ((CORE_ADDR) (cproc.new.context
					       + MACHINE_CPROC_PC_OFFSET),
				  (char *) &gdb_thread->pc,
				  sizeof(gdb_thread->pc)))
	      gdb_thread->sp = cproc.new.context + MACHINE_CPROC_SP_OFFSET;
	    else
	      {
		gdb_thread->fp = 0;
		gdb_thread->pc = 0;
	      }
	  cproc_list = (CORE_ADDR) cproc.new.list;
	}
      else
	{
	  gdb_thread->state.cthread = cproc.old.state;
	  gdb_thread->wired = TRUE;
	  gdb_thread->stack_base = cproc.old.stack_base;
	  gdb_thread->stack_top = cproc.old.stack_base + cproc.old.stack_size;
	  cproc_list = (CORE_ADDR) cproc.old.list;
	}

      if (cproc.new.incarnation)	/* Same place both versions. */
	{
	  if (inferior_read ((CORE_ADDR) cproc.new.incarnation,
			     (char *) &cthread, cthread_size))
	    {
	      unsigned int  i;
	      char *	    ptr;

	      if (new_threads)
	        ptr = (char *) cthread.new.name;
	      else
	        ptr = cthread.old.name;
	      if (ptr && inferior_read((CORE_ADDR) ptr, name, sizeof(name)))
	        {
		  gdb_thread->name = ptr = xmalloc (sizeof(name));
		  for (i = 0; i < (sizeof(name) - 1) && name[i]; i++)
		    ptr[i] = name[i];
		  ptr[i] = '\0';
		  (void) make_cleanup (free_current_contents,&gdb_thread->name);
		}
	      else if (ptr)
		message ("Unable to read name for C-Thread %u.", thread_count);
	    }
	  else
	    message ("Unable to read part of a C-Thread, continuing...");
	}
      thread_count += 1;
    }

  discard_cleanups (old_cleanup);

  *thread_count_ptr = thread_count;
  return thread_list;
}

private void
inferior_init ()
{
  kern_return_t kr;

  inferior.mid = map_port_name_to_mid (inferior.task, MACH_TYPE_TASK);

  kr = mach_port_allocate (mach_task_self (), MACH_PORT_RIGHT_RECEIVE,
			   &inferior.notify_port);
  if (kr != KERN_SUCCESS)
    fatal ("Unable to create notify port: %s.", mach_error_string(kr));

  kr = mach_port_move_member (mach_task_self(), inferior.notify_port,
			      gdb_port_set);
  if (kr != KERN_SUCCESS)
    fatal ("Unable to add notify port to port set: %s.", mach_error_string(kr));

  /*
   * Make sure that we are notified when the task dies.
   */
  notify_request (inferior.task, MACH_TYPE_TASK);

  kr = mach_port_allocate (mach_task_self(), MACH_PORT_RIGHT_RECEIVE,
			   &inferior.exception_port);
  if (kr != KERN_SUCCESS)
    fatal ("Unable to allocate task exception port: %s.",
	   mach_error_string (kr));

  /*
   * Add send right, because task_set_exception_port needs it.
   */
  kr = mach_port_insert_right (mach_task_self (), inferior.exception_port,
			       inferior.exception_port,MACH_MSG_TYPE_MAKE_SEND);
  if (kr != KERN_SUCCESS)
    fatal ("Unable to add send-right for task exception port: %s.",
	   mach_error_string (kr));

  kr = mach_port_move_member (mach_task_self(), inferior.exception_port,
			      gdb_port_set);
  if (kr != KERN_SUCCESS)
    fatal ("Unalbe to add exception port to port set: %s.",
	   mach_error_string (kr));

  kr = task_get_exception_port (inferior.task, &inferior.old_exception_port);
  if (kr != KERN_SUCCESS)    
    fatal ("Unable to get old task exception port: %s.", mach_error_string(kr));

  kr = task_set_exception_port (inferior.task, inferior.exception_port);
  if (kr != KERN_SUCCESS)    
    fatal ("Unable to set task exception port: %s.", mach_error_string(kr));

  kr = mach_port_deallocate (mach_task_self (), inferior.exception_port);
  if (kr != KERN_SUCCESS)
    fatal ("Unable to deallocate exception port send right: %s.",
	   mach_error_string (kr));

  inferior.in_setup = FALSE;

  inferior.emulator_present = have_emulator_p (inferior.task);
}

private void
inferior_exception_discard ()
{
  exc_rec_t *	cur_exc = inferior.exc_list;

  if (cur_exc == NULL)
    return;

  inferior.exc_list = cur_exc->next;
  if (MACH_PORT_VALID (cur_exc->thread))
    (void) mach_port_deallocate (mach_task_self(), cur_exc->thread);
  free (cur_exc);
}


/*
 * Set up GDB to handle the next exception in the INFERIOR's exception list.
 */
private exc_rec_t *
inferior_exception_next ()
{
  kern_return_t	kr;
  exc_rec_t *	exc_rec;
  gdb_thread_t *gdb_thread;
  mach_port_t	old_thread;

  exc_rec = inferior.exc_list;
  while (exc_rec)
    {
      if (exc_rec->mid != inferior.mid)
	gdb_thread = thread_lookup_by_mid (exc_rec->mid);
      else
	{
	  if (MACH_PORT_VALID (exc_rec->thread))
	    gdb_thread = thread_lookup (exc_rec->thread);
	  else
	    gdb_thread = NULL;

	  if (gdb_thread)
	    {
	      /*
	       * This exception was raised by a thread that wasn't known
	       * to GDB at the time the exception was caught.
	       */
	      exc_rec->mid = gdb_thread->mid;
	    }
	  else if (exc_rec->exception == EXC_SOFTWARE
		   && (exc_rec->code == EXC_GDB_STOPPED
		       || exc_rec->code == EXC_GDB_EXITED
		       || exc_rec->code == EXC_GDB_SIGNALED))
	    {
	      /*
	       * This is a task exception (U*x signal).  If the task
	       * is dead, then this is probably its exit status.  If it
	       * still alive then select the current thread if it still
	       * exists, otherwise use the first thread if there is one.
	       */
	      if (MACH_PORT_VALID (inferior.task))
		{
		  gdb_thread = thread_lookup (current_thread);
		  if (gdb_thread == NULL && inferior.thread_count > 0)
		    gdb_thread = &inferior.thread_list[0];
		}
	      else
		break;
	    }
	}

      if (gdb_thread)
	{
	  old_thread = current_thread;
	  kr = thread_setup (gdb_thread->mid);
	  if (kr == KERN_SUCCESS)
	    {
	      if (current_thread != old_thread)
		message ("Gdb selected thread %u.", gdb_thread->mid);
	      break;
	    }
	}

      message ("Thread %u disappeared after raising an exception.",
	       exc_rec->mid);
      print_exception ("Missing thread", exc_rec->exception, exc_rec->code,
		       exc_rec->sub_code);
      if (exc_rec->exception == EXC_SOFTWARE
	  && (exc_rec->code == EXC_GDB_STOPPED
	      || exc_rec->code == EXC_GDB_EXITED
	      || exc_rec->code == EXC_GDB_SIGNALED))
	inferior.stopped_in_server = TRUE;
      inferior_exception_discard ();
      exc_rec = inferior.exc_list;
    }

  if (exc_rec && exc_rec->exception == EXC_SOFTWARE
      && (exc_rec->code == EXC_GDB_STOPPED || exc_rec->code == EXC_GDB_EXITED
	  || exc_rec->code == EXC_GDB_SIGNALED))
    inferior.stopped_in_server = TRUE;

  return exc_rec;
}

/*
 * This is done in GDB after doing the FORK, but before calling
 * STARTUP_INFERIOR.
 */

private void
inferior_prepare (pid)
     int pid;
{
  push_target (&mach3_ops);

  inferior.task = task_by_pid (pid);

  if (! MACH_PORT_VALID (inferior.task))
    {
      inferior.task = MACH_PORT_NULL;
      unpush_target (&mach3_ops);
      error ("Can not map Unix pid %d to Mach task.", pid);
    }

  /*
   * Initialize the inferior state for debugging.
   */
  inferior_init ();

  inferior.in_setup = TRUE;
}

private boolean_t
inferior_process_startup_msg (in_msg, out_msg)
     mach_msg_header_t *in_msg;
     mach_msg_header_t *out_msg;
{
  boolean_t	ux_exec_trap_seen = FALSE;

  if (notify_server (in_msg, out_msg))
    {
      if (inferior.task == MACH_PORT_NULL)
	mach3_mourn_inferior ();
      else
	mach3_kill ();
      error ("Received a notification while intercepting syscalls.");
    }
  else if (gdb_exception_server ((exception_msg_t *) in_msg,
				 (exception_reply_t *) out_msg))
    {
      exc_rec_t *exc_rec = inferior.exc_list;

      if (exc_rec)
	if (exc_rec->exception == EXC_SOFTWARE
	    && exc_rec->code == EXC_GDB_STOPPED
	    && (exc_rec->sub_code == SIGTRAP || exc_rec->sub_code == SIGCHLD))
	  {
	    /*
	     * Assume that U*x tracing is working and that this
	     * is an EXEC trap for a direct child or GDB's child (the shell)
	     * is creating sub-processes (processing a .cshrc or .profile).
	     */
	    if (exc_rec->sub_code == SIGTRAP)
	      {
		ux_exec_trap_seen = TRUE;
		exc_rec->sub_code = 0;
	      }
	    inferior.stopped_in_server = TRUE;
	    inferior_resume (exc_rec->sub_code);
	  }
	else
	  {
	    message ("Inferior raised exception during set up.");
	    print_exception ("Inferior", exc_rec->exception, exc_rec->code,
			     exc_rec->sub_code);
	    error ("You can try examinging the current state, but no promises that it will work.");
	  }
    }
  else if (! gdb_message_server (in_msg))
    message ("Ignoring message %d during set up.", in_msg->msgh_id);

  return ux_exec_trap_seen;
}

/*
 * Read inferior task's LEN bytes from ADDR and copy it to MYADDR
 * in gdb's address space.
 *
 * Return 0 on failure; number of bytes read otherwise.
 */
private int
inferior_read (addr, myaddr, length)
     CORE_ADDR addr;
     char *myaddr;
     int length;
{
  kern_return_t kr;
  vm_address_t low_address = (vm_address_t) trunc_page (addr);
  vm_size_t    aligned_length = 
    			(vm_size_t) round_page (addr+length) - low_address;
  vm_address_t copied_memory;
  vm_size_t    copy_count;

  /* Get memory from inferior with page aligned addresses */
  kr = vm_read (inferior.task, low_address, aligned_length, &copied_memory,
		(mach_msg_type_number_t *) &copy_count);
  if (kr != KERN_SUCCESS)
    {
      inferior_check ();
      return 0;
    }

  memcpy (myaddr, (char *)addr - low_address + copied_memory, length);

  kr = vm_deallocate (mach_task_self (), copied_memory, copy_count);
  if (kr != KERN_SUCCESS)
    message ("inferior_read: vm_deallocate: %s", mach_error_string (kr));

  return length;
}

/*
 * Restore the state of the inferior task so it can be detached.
 */
private void
inferior_restore (signal)
     int	signal;
{
  kern_return_t	kr;
  unsigned int	i;

#ifdef	PT_DETACH
  if (inferior.has_ptrace_server)
    {
      if (! inferior.stopped_in_server)
	kill (inferior_pid, SIGSTOP);

      kr = our_ptrace (PT_DETACH, signal);
      if (kr != KERN_SUCCESS)
	message ("Error disabling U*x tracing: %s.", mach_error_string (kr));

      /*
       * This will keep inferior_resume() from confusing the server.
       */
      inferior.stopped_in_server = FALSE;
    }
#else	/* not defined(PT_DETACH) */
  if (inferior.has_ptrace_server
      && ! query ("GDB cannot disable U*x tracing.  Detach anyway? "))
    error ("Inferior not detached.");
#endif

  kr = task_set_exception_port (inferior.task, inferior.old_exception_port);
  if (kr != KERN_SUCCESS)
    message ("Could not reset the inferior's exception port: %s",
	     mach_error_string (kr));

  if (remove_breakpoints ())
    message ("Could not remove breakpoints when detaching.");

  for (i = 0; i < inferior.thread_count; i++)
    thread_restore (&inferior.thread_list[i]);

}

/*
 * Resume the inferior task.
 */
private void
inferior_resume (send_signal)
     int send_signal;
{
  kern_return_t kr;

  if (inferior.stopped_in_server)
    {
      inferior.stopped_in_server = FALSE;
      kr = our_ptrace (PT_CONTINUE, send_signal);
      if (kr != KERN_SUCCESS)
	message ("U*x continue failed: %s.", mach_error_string (kr));
    }
  
  if (inferior.suspended)
    {
      kr = task_resume (inferior.task);
      if (kr != KERN_SUCCESS)
	{
	  if (kr == KERN_FAILURE)
	    message ("Task was not suspended!");
	  else
	    {
	      mach3_mourn_inferior ();
	      if (kr == KERN_INVALID_ARGUMENT)
		error ("The inferior task died.");
	      else
		error ("inferior_resume: task_resume strange error: %s",
		       mach_error_string (kr));
	    }
	}
      inferior.suspended = FALSE;
    }

  /*
   * Discard the first exception record, leaving the rest for mach3_wait()
   * to handle.
   */
  if (inferior.exc_list)
    inferior_exception_discard ();
}

/*
 * Set up the inferior for debugging.
 */
private void
inferior_setup ()
{
  kern_return_t	kr;
  unsigned int	i;
  mach_msg_type_number_t thread_count;
  thread_t *	thread_list;
  thread_t	saved_thread;

  if (inferior.suspended && inferior.setup)
    return;

  kr = task_threads (inferior.task, &thread_list, &thread_count);
  if (kr != KERN_SUCCESS)
    {
      mach3_mourn_inferior ();
      error ("Inferior task found dead.");
    }

  inferior.setup = TRUE;

  if (thread_count == 0)
    {
      while (inferior.thread_count)
	thread_cleanup (inferior.thread_list[0].port);

      message ("Task %u has no threads.", inferior.mid);
      return;
    }

  if (thread_count > inferior.thread_count)
    inferior.thread_list = (gdb_thread_t *) xrealloc (inferior.thread_list,
						      thread_count *
						      sizeof(gdb_thread_t));

  saved_thread = current_thread;

  /*
   * Remove any threads that have disappeared.  Normally, this will be
   * a no-op.
   *
   * NOTE 1: Thread_cleanup() decrements inferior.thread_count.
   * 
   * NOTE 2: This assumes that the kernel gives us the threads in the
   * same order each time.  It is still correct if the order changes,
   * but it will be extremely inefficient.
   */
  for (i = 0; i < inferior.thread_count; i++)
    if (thread_list[i] == inferior.thread_list[i].port)
      thread_setup (inferior.thread_list[i].mid);
    else
      thread_cleanup (inferior.thread_list[i--].port);

  /*
   * Initialize our thread information, including a port reference.
   * 
   * NOTE: thread_init() will increment inferior.thread_count.
   */
  for (/* CONTINUE */; i < thread_count; i++)
    {
      thread_init (thread_list[i]);
      thread_list[i] = MACH_PORT_NULL;
    }

  if (saved_thread)
    {
      if (saved_thread != current_thread)
	{
	  gdb_thread_t *gdb_thread = thread_lookup (saved_thread);

	  if (gdb_thread)
	    thread_setup (gdb_thread->mid);
	  else
	    {
	      message ("Current thread disappeared.");
	      registers_changed ();
	      current_thread = MACH_PORT_NULL;
	    }
	}
    }
  else
    {
      registers_changed ();
      current_thread = MACH_PORT_NULL;
    }

  consume_thread_list (thread_list, thread_count);

  inferior_setup_cthreads ();

  return;
}

/*
 * Fetch the inferior's cthreads and set up GDB's data structures
 */
private void
inferior_setup_cthreads ()
{
  gdb_thread_t *kthread;
  gdb_thread_t *thread_list;
  unsigned int	thread_count;
  unsigned int	no_kthread_count;
  unsigned int	i;
  unsigned int	k;

  thread_count = inferior.thread_count + inferior.cthread_count;
  thread_list = inferior_get_cthreads (&thread_count);

  if (thread_list == NULL)
    {
      inferior.cthread_count = 0;
      if (inferior.cthread_list)
	{
	  for (i = 0; i < inferior.cthread_count; i++)
	    if (inferior.cthread_list[i].name)
	      free (inferior.cthread_list[i].name);
	  free (inferior.cthread_list);
	  inferior.cthread_list = NULL;
	}
      return;
    }

  for (k = 0; k < inferior.thread_count; k++)
    {
      kthread = &inferior.thread_list[k];

      if (kthread->name)
	{
	  free (kthread->name);
	  kthread->name = NULL;
	}
      kthread->state.swapped = FALSE;
      kthread->state.has_cthread = FALSE;
      kthread->state.cthread = 0;
      kthread->wired = FALSE;
      kthread->stack_base = kthread->stack_top = 0;
    }

  no_kthread_count = 0;
  for (i = 0; i < thread_count; i++)
    {
      gdb_thread_t *	cthread = &thread_list[i];

      for (k = 0; k < inferior.thread_count; k++)
        {
	  CORE_ADDR	sp;

	  kthread = &inferior.thread_list[k];

	  if (kthread->in_emulator)
	    sp = kthread->user_sp;
	  else
	    sp = kthread->sp;

	  if (cthread->stack_base < sp && sp <= cthread->stack_top)
	    {
	      kthread->name = cthread->name;
	      kthread->state.has_cthread = TRUE;
	      kthread->state.cthread = cthread->state.cthread;
	      kthread->wired = cthread->wired;
	      kthread->stack_base = cthread->stack_base;
	      kthread->stack_top = cthread->stack_top;
	      /*
	       * Mark this cthread as saved.
	       */
	      cthread->stack_base = cthread->stack_top;
	      break;
	    }
	}

      if (cthread->stack_base != cthread->stack_top)
        {
	  /*
	   * If we have seen this thread before save its MID.
	   */
	  for (k = 0; k < inferior.cthread_count; k++)
	    if (inferior.cthread_list[k].stack_base == cthread->stack_base)
	      {
		cthread->mid = inferior.cthread_list[k].mid;
		break;
	      }
	  no_kthread_count += 1;
	}
    }

  if (inferior.cthread_list)
    {
      for (i = 0; i < inferior.cthread_count; i++)
        if (inferior.cthread_list[i].name)
	  free (inferior.cthread_list[i].name);
      free (inferior.cthread_list);
      inferior.cthread_list = NULL;
    }

  /*
   * Save the un-matched C-threads.
   */
  inferior.cthread_count = no_kthread_count;
  if (no_kthread_count)
    {
      inferior.cthread_list = (gdb_thread_t *) xmalloc (no_kthread_count *
							sizeof(gdb_thread_t));
      k = 0;
      for (i = 0; i < thread_count; i++)
	if (thread_list[i].stack_base != thread_list[i].stack_top)
	  {
	    if (thread_list[i].mid == 0)
	      thread_list[i].mid = ++inferior.mid_counter;
	    if (thread_list[i].name == NULL)
	      {
		/*
		 * Allocate enough space to print the prefix and the id.
		 */
		thread_list[i].name = xmalloc (sizeof(k) * DIGITS_PER_CHAR
					       + sizeof(C_THREAD_PREFIX));
		sprintf(thread_list[i].name, "%s%d", C_THREAD_PREFIX, k);
	      }
	    inferior.cthread_list[k++] = thread_list[i];
	  }
    }
  free (thread_list);

  for (k = 0; k < inferior.thread_count; k++)
    if (inferior.thread_list[k].name == NULL)
      {
	/*
	 * Allocate enough space to print the prefix and the id.
	 */
	inferior.thread_list[k].name = xmalloc (sizeof(k) * DIGITS_PER_CHAR
						+ sizeof(K_THREAD_PREFIX));
	sprintf(inferior.thread_list[k].name, "%s%d", K_THREAD_PREFIX, k);
      }
}

/*
 * Called from inferior after FORK but before EXEC.
 */
private void
inferior_setup_self ()
{
  kern_return_t kr;
  
  /*
   * Try and turn on SIGNAL forwarding.
   */
  if (our_ptrace (PT_TRACE_ME, 0) != KERN_SUCCESS)
    message ("[Inferior]: Unable to turn on tracing.", mach_error_string (kr));

  /*
   * Get the NAME of the bootstrap port in this task so that GDB can read it.
   */
  kr = task_get_bootstrap_port (mach_task_self (), &gdb_ux_server_port_name);
  if (kr != KERN_SUCCESS)
    abort ();

  /*
   * Don't deallocate port in order to minimize the chance of the port name
   * being re-used.
   */

  /*
   * Suspend this task so that GDB can interpose on the bootstrap port.
   * We will be resumed by GDB.
   */
  (void) task_suspend (mach_task_self ());
}

private kern_return_t
inferior_suspend ()
{
  kern_return_t	kr = KERN_SUCCESS;
  unsigned int	i;

  if (MACH_PORT_VALID (inferior.task) && ! inferior.suspended)
    {
      inferior.suspended = TRUE;
      inferior.setup = FALSE;

      kr = task_suspend (inferior.task);

      /*
       * Make sure that we will call abort on any runnable threads.
       *
       * NOTE:  This is not perfect.  If the task is suspending/resuming
       * its threads we may have problems.
       */
      for (i = 0; i < inferior.thread_count; i++)
	if (! inferior.thread_list[i].suspended)
	  inferior.thread_list[i].aborted = FALSE;
    }

  return kr;
}

/*
 * Write inferior task's LEN bytes from ADDR and copy it to MYADDR
 * in gdb's address space.
 */
private int
inferior_write (addr, myaddr, length)
     CORE_ADDR addr;
     char *myaddr;
     int length;
{
  kern_return_t kr;
  vm_address_t	low_address    = (vm_address_t) trunc_page (addr);
  vm_size_t	aligned_length = 
    			(vm_size_t) round_page (addr+length) - low_address;
  vm_address_t	copied_memory;
  vm_size_t	copy_count;
  boolean_t	change_protection = FALSE;
#ifdef	FLUSH_INFERIOR_CACHE
  boolean_t	is_executable = FALSE;
#endif
  vm_region_list_t *region_element;
  vm_region_list_t *region_head = (vm_region_list_t *)NULL;
  vm_region_list_t *scan;
  vm_size_t	remaining_length;
  vm_address_t	region_address;

  /* Get memory from inferior with page aligned addresses */
  kr = vm_read (inferior.task, low_address, aligned_length, &copied_memory,
		 &copy_count);
  if (kr != KERN_SUCCESS)
    {
      inferior_check ();
      message ("write_inferior: vm_read: %s", mach_error_string (kr));
      return 0;
    }

  memcpy ((char *)addr - low_address + copied_memory, myaddr, length);

  obstack_init (&region_obstack);

  /* Do writes atomically.
   * First check for holes and unwritable memory.
   */
  remaining_length  = aligned_length;
  region_address    = low_address;
  while(region_address < low_address + aligned_length)
    {
      vm_prot_t protection;
      vm_prot_t max_protection;
      vm_inherit_t inheritance;
      boolean_t shared;
      mach_port_t object_name;
      vm_offset_t offset;
      vm_size_t   region_length = remaining_length;
      vm_address_t old_address  = region_address;

      kr = vm_region (inferior.task, &region_address, &region_length,
		      &protection, &max_protection, &inheritance, &shared,
		      &object_name, &offset);
      if (kr != KERN_SUCCESS)
	{
	  message ("write_inferior: vm_region: %s. Nothing written.",
		   mach_error_string (kr));
	  length = 0;
	  goto out;
	}

      /* Check for holes in memory */
      if (old_address != region_address)
	{
	  message ("No memory at 0x%x. Nothing written.", old_address);
	  length = 0;
	  goto out;
	}

      if ((max_protection & VM_PROT_WRITE) == 0)
	{
	  message ("Memory at address 0x%x is unwritable. Nothing written.",
		   old_address);
	  length = 0;
	  goto out;
	}

      if ((protection & VM_PROT_WRITE) == 0)
	change_protection = TRUE;

#ifdef	FLUSH_INFERIOR_CACHE
      /*
       * Do we need to flush this data from the I-Cache?
       */
      if (protection & VM_PROT_EXECUTE)
	is_executable = TRUE;
#endif

      /* Chain the regions for later use */
      region_element = 
	(struct vm_region_list *)
	  obstack_alloc (&region_obstack, sizeof (struct vm_region_list));

      region_element->protection = protection;
      region_element->start      = region_address;
      region_element->length     = region_length;

      /* Chain the regions along with protections */
      region_element->next = region_head;
      region_head          = region_element;

      region_address += region_length;
      remaining_length = remaining_length - region_length;
    }

  /*
   * If things fail after this, we give up.
   * Somebody is messing up inferior.task's mappings.
   */

  /* Enable writes to the chained vm regions */
  if (change_protection)
    for (scan = region_head; scan; scan = scan->next)
      if (!(scan->protection & VM_PROT_WRITE))
	{
	  kr = vm_protect (inferior.task, scan->start, scan->length, FALSE,
			   (scan->protection | VM_PROT_WRITE));
	  if (kr != KERN_SUCCESS)
	    {
	      message ("vm_protect(write-enable): %s", mach_error_string (kr));
	      length = 0;
	      goto out;
	    }
	}

  kr = vm_write (inferior.task, low_address, copied_memory, aligned_length);
  if (kr != KERN_SUCCESS)
    {
      message ("vm_write failed: %s.  No bytes written.",
	       mach_error_string (kr));
      length = 0;
    }

 out:

  /*
   * Restore the original region protections if they were changed.
   */
  if (change_protection)
    for (scan = region_head; scan; scan = scan->next)
      {
	boolean_t first = TRUE;

	if (!(scan->protection & VM_PROT_WRITE))
	  {
	    kr = vm_protect (inferior.task, scan->start, scan->length, FALSE,
			     scan->protection);
	    if (kr != KERN_SUCCESS && first)
	      {
		message ("vm_protect(write-disable): %s",
			 mach_error_string (kr));
		first = FALSE;
	      }
	  }
      }

#ifdef	FLUSH_INFERIOR_CACHE
  /*
   * When over-writing code on some machines the I-Cache must be flushed
   * explicitly.
   */
  if (is_executable)
    {
      vm_machine_attribute_val_t	flush = MATTR_VAL_ICACHE_FLUSH;

      kr = vm_machine_attribute (inferior.task, (vm_address_t) addr,
				 (vm_size_t) length,  MATTR_CACHE, &flush);
      if (kr != KERN_SUCCESS)
        message ("Unable to flush I-Cache (%d bytes at 0x%x): %s.",
		 length, addr, mach_error_string (kr));
    }
#endif	/* defined(FLUSH_INFERIOR_CACHE) */

  obstack_free (&region_obstack, 0);

  (void) vm_deallocate (mach_task_self (), copied_memory, copy_count);

  return length;
}

/*
 * Returns the address of variable NAME or 0 if not found.
 */
private CORE_ADDR
lookup_address_of_variable (name)
     char *name;
{
  struct symbol *	sym;
  CORE_ADDR		symaddr = 0;
  struct minimal_symbol *msymbol;

  sym = lookup_symbol (name, (struct block *) NULL, VAR_NAMESPACE, (int *) NULL,
		       (struct symtab **) NULL);

  if (sym)
    symaddr = SYMBOL_VALUE_ADDRESS (sym);

  if (symaddr == 0)
    {
      msymbol = lookup_minimal_symbol (name, (struct objfile *) NULL);

      if (msymbol && MSYMBOL_TYPE (msymbol) == mst_data)
	symaddr = SYMBOL_VALUE_ADDRESS (msymbol);
    }

  return symaddr;
}

/*
 * Format a thread id.
 *
 * This is called using the hook THREAD_OUTPUT_ID from breakpoint.c.
 */
char *
mach3_output_thread_id (mid)
     mach_id_t mid;
{
  static char foobar [20];

  if (mid > 0)
    {
      sprintf (foobar, "mid %u", mid);
      return foobar;
    }
  else
    return "*any thread*";
}

/*
 * Map a thread id string (MID or a @SLOTNUMBER) to a thread-id.
 *
 * A NULL or blank argument returns 0.  The meaning of 0, which is
 * an invalid Mach id, is left up to the caller.  Known interpretations
 * are: the first kernel thread, all threads, and invalid.
 *
 * Otherwise it returns a valid Mach id or calls ERROR.
 *
 * This is called using the hook THREAD_PARSE_ID from breakpoint.c.
 */
mach_id_t
mach3_parse_thread_id (arg)
     char *arg;
{
  mach_id_t mid;
  
  if (arg == 0)
    return 0;
  
  while (*arg == ' ' || *arg == '\t')
    arg++;
  
  if (*arg == '\0')
    return 0;
  
  if (isdigit (*arg))
    {
      mid = atoi (arg);
      if (mid == 0)
	error ("Zero is not a valid MID.");
      return mid;
    }
  
  if (*arg == '@')
    arg++;

  if (! isdigit (*arg))
    error ("Invalid thread id.  Use either MID or @SLOTNUMBER.");

  mid = atoi (arg);

  if (mid >= inferior.thread_count)
    error ("Invalid slot number %u > %u.", mid, inferior.thread_count);

  return inferior.thread_list[mid].mid;
}

/*
 * If we are stopped because of a breakpoint exception and we have
 * switched threads, we need to switch back so that GDB can step over
 * the breakpoint.  Returns 1 if we have switched threads, otherwise 0.
 * 
 * This routine is called using the hook PREPARE_TO_PROCEED() from infrun.c.
 */

int
mach3_prepare_to_proceed ()
{
  mach_id_t	mid = map_port_name_to_mid (current_thread, MACH_TYPE_THREAD);

  if (inferior.exc_list && mid != inferior.exc_list->mid
      && inferior.exc_list->exception == EXC_BREAKPOINT
      && thread_setup (inferior.exc_list->mid) == KERN_SUCCESS)
    return 1;

  return 0;
}

/*
 * Handle a ^C typed by the user.
 *
 * If we are the main thread and it is safe to quit immediately,
 * then we return to the top-level.
 *
 * If we are the "signal" thread then we can only pass the signal
 * on the main thread.  If it is blocked in a mach_msg wait this
 * requires sending it a message, otherwise we expect it to
 * notice the quit_flag, eventually.
 *
 * Called by REQUEST_QUIT() from utils.c(request_quit)
 */

void
mach3_request_quit ()
{
  if (cthread_self () == gdb_main_thread_id)
    {
      if (immediate_quit)
	quit ();
    }
  else
    {
      /*
       * We are not the main thread so we cannot longjmp back to the
       * top-level prompt.  Pass the signal onto the main thread if
       * possible.
       */
      mutex_lock (&gdb_quit_lock);
      if (gdb_really_waiting)
        wakeup_main_thread ();
      mutex_unlock (&gdb_quit_lock);
    }
}      

/*
 * Intercept system calls to Unix server.
 * After EXEC_COUNTER calls to exec(), return.
 *
 * Pre-assertion:  Child is suspended. (Not verified)
 * Post-condition: Child is suspended after EXEC_COUNTER exec() calls.
 *
 * Called by STARTUP_INFERIOR hook from fork-child.c
 */

void
mach3_startup_inferior (exec_counter)
     int exec_counter;
{
  struct syscall_msg_t {
    mach_msg_header_t	header;
    mach_msg_type_t 	type;
    char room[ 2000 ];	/* Enuff space */
  };

  struct syscall_msg_t syscall_in, syscall_out;
  int		ux_exec_counter;
  boolean_t	ux_exec_trap_seen;
  mach_port_t	fake_server;
  mach_port_t	gdb_ux_server_send;
  mach_port_t	original_exec_reply;
  mach_port_t	exec_reply;
  boolean_t	set_header;
  mach_msg_type_name_t acquired;
  struct task_basic_info info;
  mach_msg_type_number_t info_count;
  mach_port_urefs_t refs;

  kern_return_t kr;

  if (exec_counter <= 0)
    return;		/* We are already set up in the correct program */

  /*
   * Wait for inferior task to suspend itself.
   */
  while(1)
    {
      info_count = sizeof(info);
      kr = task_info (inferior.task, TASK_BASIC_INFO, (task_info_t) &info,
		      &info_count);
      if (kr != KERN_SUCCESS)
	if (kr == KERN_INVALID_ARGUMENT)
	  {
	    mach3_mourn_inferior ();
	    error ("Inferior Task died.");
	  }
	else
	  fatal ("Task_info call failed while trying to start inferior: %s.",
		 mach_error_string (kr));

      if (info.suspend_count)
	break;

      /* 
       * This will depress the thread's priority for 10ms.
       */
      (void) thread_switch (MACH_PORT_NULL, SWITCH_OPTION_DEPRESS, 10);
    }

  kr = mach_port_allocate(mach_task_self (), MACH_PORT_RIGHT_RECEIVE,
			  &fake_server);
  if (kr != KERN_SUCCESS)
    fatal ("Unable to create fake server port: %s.", mach_error_string (kr));
  
  /*
   * Read the inferior's bootstrap port name.
   */
  if (!inferior_read ((CORE_ADDR) &gdb_ux_server_port_name,
		      (char *) &gdb_ux_server_port_name,
		      sizeof(gdb_ux_server_port_name)))
    {
      mach3_kill ();
      error ("Can't read inferior task bootstrap port name.");
    }

  kr = mach_port_get_refs (inferior.task, gdb_ux_server_port_name,
			   MACH_PORT_RIGHT_SEND, &refs);
  if (kr != KERN_SUCCESS)
    {
      mach3_kill ();
      error ("Can't get references for inferior bootstrap port: %s.",
	     mach_error_string (kr));
    }

  if (refs > 1)
    {
      kr = mach_port_mod_refs (inferior.task, gdb_ux_server_port_name,
			       MACH_PORT_RIGHT_SEND, -(refs - 1));
      if (kr != KERN_SUCCESS)
        {
	  mach3_kill ();
	  error ("Can't decrement references for inferior bootstrap port: %s.",
		 mach_error_string (kr));
	}
    }

  kr = mach_port_extract_right (inferior.task, gdb_ux_server_port_name,
				MACH_MSG_TYPE_MOVE_SEND, &gdb_ux_server_send,
				&acquired);
  if (kr != KERN_SUCCESS || acquired != MACH_MSG_TYPE_PORT_SEND)
    {
      mach3_kill ();
      error ("Unable to extract inferior bootstrap port: %s.",
	     mach_error_string (kr));
    }

  kr = mach_port_insert_right (inferior.task, gdb_ux_server_port_name,
			       fake_server, MACH_MSG_TYPE_MAKE_SEND);
  if (kr != KERN_SUCCESS)
    {
      mach3_kill ();
      error ("Unable to insert fake bootstrap port: %s.",
	     mach_error_string (kr));
    }

  xx_debug ("Inferior task server ports set up: fs %x, ospn %x, oss %x.\n",
	    fake_server, gdb_ux_server_port_name, gdb_ux_server_send);

  if (refs > 1)
    {
      kr = mach_port_mod_refs (inferior.task, gdb_ux_server_port_name,
			       MACH_PORT_RIGHT_SEND, refs-1);
      if (kr != KERN_SUCCESS)
	{
	  mach3_kill ();
	  error ("Can't increment references for fake bootstrap port: %s.",
		 mach_error_string (kr));
	}
    }

  /*
   * A receive right to the reply generated by unix server exec() request.
   */
  kr = mach_port_allocate(mach_task_self (), MACH_PORT_RIGHT_RECEIVE,
			  &exec_reply);
  if (kr != KERN_SUCCESS)
    fatal ("Unable to allocate EXEC reply port: %s.", mach_error_string (kr));
    
  kr = mach_port_move_member(mach_task_self(), fake_server, gdb_port_set);
  if (kr != KERN_SUCCESS)
    fatal ("Unable to add fake bootstrap port to port set: %s.",
	   mach_error_string (kr));

  xx_debug ("Syscall fake server set up, resuming inferior.\n");
  
  kr = task_resume (inferior.task);
  if (kr != KERN_SUCCESS)
    {
      mach3_mourn_inferior ();
      error ("Inferior task died.");
    }

#define	SET_HEADER(bits, local)						\
    (bits = (MACH_MSGH_BITS_OTHER(bits)					\
	     | MACH_MSGH_BITS(MACH_MSG_TYPE_COPY_SEND, local)))
  /*
   * Read requests from the inferior.
   * Pass directly through everything else except exec() calls.
   */
  ux_exec_counter = exec_counter;
  ux_exec_trap_seen = FALSE;
  while (exec_counter > 0)
    {
      kr = mach_msg (&syscall_in.header, MACH_RCV_MSG, 0,
		     sizeof(struct syscall_msg_t), gdb_port_set,
		     MACH_MSG_TIMEOUT_NONE, MACH_PORT_NULL);
      if (kr != KERN_SUCCESS)
        {
	  mach3_kill ();
	  error ("Error receiving intercepted syscall: %s.",
		 mach_error_string (kr));
	}
	    
#ifdef DUMP_SYSCALL
      print_msg (&syscall_in.header);
#endif

      /* ASSERT : msgh_local_port == fake_server */

      if (syscall_in.header.msgh_local_port != fake_server)
	{
	  if (inferior_process_startup_msg (&syscall_in.header,
					    &syscall_out.header))
	    {
	      ux_exec_trap_seen = TRUE;
	      ux_exec_counter -= 1;
	    }
	  continue;
	}

      set_header = TRUE;
      if (syscall_in.header.msgh_id == MIG_EXEC_SYSCALL_ID)
	{
	  xx_debug ("Received EXEC SYSCALL, counter = %d\n", exec_counter);
	  if (exec_counter-- == 1)
	    {
	      original_exec_reply = syscall_in.header.msgh_remote_port;
	      syscall_in.header.msgh_local_port = exec_reply;
	      SET_HEADER (syscall_in.header.msgh_bits,
			  MACH_MSG_TYPE_MAKE_SEND_ONCE);
	      set_header = FALSE;
	    }
	}

      if (set_header)
        {
	  syscall_in.header.msgh_local_port =syscall_in.header.msgh_remote_port;
	  SET_HEADER (syscall_in.header.msgh_bits,
		      MACH_MSGH_BITS_REMOTE(syscall_in.header.msgh_bits));
	}

      syscall_in.header.msgh_remote_port = gdb_ux_server_send;

      kr = mach_msg_send (&syscall_in.header);
      if (kr != KERN_SUCCESS)
        {
	  mach3_kill ();
	  error ("Unable to forward syscall to original server: %s.",
		 mach_error_string (kr));
	}
    }
	
  kr = mach_port_move_member(mach_task_self(), fake_server, MACH_PORT_NULL);
  if (kr != KERN_SUCCESS)
    fatal ("Unable to remove fake bootstrap port from port set: %s.",
	   mach_error_string (kr));

  kr = mach_port_move_member(mach_task_self(), exec_reply, gdb_port_set);
  if (kr != KERN_SUCCESS)
    fatal ("Unable to add exec reply port to port set: %s.",
	   mach_error_string (kr));

  /*
   * Now wait for the exec reply message.
   */
  do {
    kr = mach_msg (&syscall_in.header, MACH_RCV_MSG, 0,
		   sizeof (struct syscall_msg_t), gdb_port_set,
		   MACH_MSG_TIMEOUT_NONE, MACH_PORT_NULL);
    if (kr != KERN_SUCCESS)
      {
	mach3_kill ();
	error ("Error receiving exec reply message: %s.",
	       mach_error_string (kr));
      }

    if (syscall_in.header.msgh_local_port != exec_reply
	&& inferior_process_startup_msg (&syscall_in.header,
					 &syscall_out.header))
      {
	ux_exec_trap_seen = TRUE;
	ux_exec_counter -= 1;
      }
  } while (syscall_in.header.msgh_local_port != exec_reply);

  kr = inferior_suspend ();
  if (kr != KERN_SUCCESS)
    {
      mach3_mourn_inferior ();
      error ("Inferior task died.");
    }

  xx_debug ("Received exec reply from bsd server, suspended inferior task.\n");

#ifdef DUMP_SYSCALL
      print_msg (&syscall_in.header);
#endif

  /*
   * Message should appear as if it came from the unix server, and
   * go to the inferior task's original reply port.
   */
  syscall_in.header.msgh_local_port = MACH_PORT_NULL;
  syscall_in.header.msgh_remote_port = original_exec_reply;
  syscall_in.header.msgh_bits =MACH_MSGH_BITS_OTHER(syscall_in.header.msgh_bits)
			      | MACH_MSGH_BITS(MACH_MSG_TYPE_MOVE_SEND_ONCE, 0);

  kr = mach_msg_send (&syscall_in.header);
  if (kr != KERN_SUCCESS)
    {
      mach3_kill ();
      error ("Unable to forward exec reply to inferior, assumed dead: %s.",
	     mach_error_string (kr));
    }

  kr = mach_port_get_refs (inferior.task, gdb_ux_server_port_name,
			   MACH_PORT_RIGHT_SEND, &refs);
  if (kr != KERN_SUCCESS)
    {
      mach3_kill ();
      error ("Can't get references for fake bootstrap port: %s.",
	     mach_error_string (kr));
    }

  kr = mach_port_mod_refs (inferior.task, gdb_ux_server_port_name,
			   MACH_PORT_RIGHT_SEND, -refs);
  if (kr != KERN_SUCCESS)
    {
      mach3_kill ();
      error ("Can't decrement references for fake bootstrap port: %s.",
	     mach_error_string (kr));
    }

  kr = mach_port_insert_right (inferior.task, gdb_ux_server_port_name,
			       gdb_ux_server_send, MACH_MSG_TYPE_MOVE_SEND);
  if (kr != KERN_SUCCESS)
    {
      mach3_kill ();
      error ("Unable to restore bootstrap port: %s.", mach_error_string (kr));
    }

  if (refs > 1)
    {
      kr = mach_port_mod_refs (inferior.task, gdb_ux_server_port_name,
			       MACH_PORT_RIGHT_SEND, refs-1);
      if (kr != KERN_SUCCESS)
        {
	  mach3_kill ();
	  error ("Can't restore references for inferior's bootstrap port: %s.",
		 mach_error_string (kr));
	}
    }

  kr = mach_port_mod_refs (mach_task_self (), fake_server,
			   MACH_PORT_RIGHT_RECEIVE, -1);
  if (kr != KERN_SUCCESS)
    fatal ("Error destroying fake bootstrap port: %s.", mach_error_string (kr));

  kr = mach_port_mod_refs (mach_task_self (), exec_reply,
			   MACH_PORT_RIGHT_RECEIVE, -1);
  if (kr != KERN_SUCCESS)
    fatal ("Error destroying fake bootstrap port: %s.", mach_error_string (kr));

  xx_debug ("Done with exec call interception\n");

  /*
   * It should now be safe to give the inferior task to the user.
   */
  inferior.in_setup = FALSE;

  /*
   * If we are getting U*x EXEC traps and we didn't see the number we
   * expected (unlikely), then warn the user.
   */
  if (ux_exec_trap_seen && ux_exec_counter != 0)
    message ("GDB received %s U*x EXEC trap exceptions it expected.",
	     (ux_exec_counter > 0 ? "fewer" : "more"));

  /*
   * This will notify the user and return to the command level if
   * the inferior task died.
   */
  inferior_setup ();

  /*
   * Set up the first kernel thread for debugging.  If there are no
   * threads, thread_setup() will print an appropriate message.
   */
  (void) thread_setup (0);

  /*
   * If the user type ^C, this is a fine place to stop.
   */
  QUIT;
}

private void *
mach3_thread (arg)
     void *arg;
{
  for (;;)
    sigpause (0);
}

/*
 * Return 1 if the MID specifies the thread that caused the last
 * exception.  Zero matches all threads.
 *
 * This is called by THREAD_ALLOWED_TO_BREAK from breakpoint.c.
 *
 * NOTE:  It also returns 1 if there are no outstanding exceptions.
 * This seemed analogous to the previous version's returning 1 when
 * it couldn't find the current thread.  What is the right thing to
 * do in that case?  -pds
 */
int
mach3_thread_for_breakpoint (mid)
     mach_id_t mid;
{
  if (mid == 0 || inferior.exc_list == NULL
      || (inferior.exc_list->exception == EXC_BREAKPOINT
	  && mid == inferior.exc_list->mid))
    return 1;

  return 0;
}

private mach_id_t
map_port_name_to_mid (name, type)
     mach_port_t name;
     mach_type_t type;
{
  port_chain_t elem;

  if (!MACH_PORT_VALID (name))
    return -1;

  elem = port_chain_lookup (notify_chain, name);

  if (elem && (elem->type == type))
    return elem->mid;
  
  if (elem)
    {
      message ("Found %x, but the type didn't match: %u != %u.",
	       name, type, elem->type);
      return 0;
    }
  
  if (! MACH_PORT_VALID (mid_server))
    {
      message ("Machid server port invalid, can not map port 0x%x to mid",
	       name);
      return 0;
    }
  else
    {
      mach_id_t mid;
      kern_return_t kr;

      kr = machid_mach_register (mid_server, mid_auth, name, type, &mid);
      
      if (kr != KERN_SUCCESS)
	{
	  message ("Can not map name (0x%x) to mid with machid", name);
	  return 0;
	}
      return mid;
    }
}

/*
 * Cancel a dead-name request.
 */
private void
notify_cancel (port)
     mach_port_t	port;
{
  port_chain_t	element;
  mach_port_t	previous_port;
  kern_return_t	kr;

  element = port_chain_lookup (notify_chain, port);
  if (element)
    {
      kr = mach_port_request_notification (mach_task_self (), port,
					   MACH_NOTIFY_DEAD_NAME, FALSE,
					   MACH_PORT_NULL,
					   MACH_MSG_TYPE_MOVE_SEND_ONCE,
					   &previous_port);

      if (kr == KERN_SUCCESS && MACH_PORT_VALID (previous_port))
	(void) mach_port_deallocate (mach_task_self (), previous_port);
      if (kr != KERN_SUCCESS && kr != KERN_INVALID_ARGUMENT)
        message ("Error canceling dead-name notification: %s.",
		 mach_error_string (kr));

      port_chain_delete (&notify_chain, port);
    }
}

/*
 * Request dead name notification.
 */
private void
notify_request (name, type)
     mach_port_t name;
     mach_type_t type;
{
  kern_return_t kr;
  mach_port_t	previous_port = MACH_PORT_NULL;
  
  if (! MACH_PORT_VALID (name))
    return;
  
  if (port_chain_lookup (notify_chain, name))
    return;

  kr = mach_port_request_notification (mach_task_self (), name,
				       MACH_NOTIFY_DEAD_NAME, TRUE,
				       inferior.notify_port,
				       MACH_MSG_TYPE_MAKE_SEND_ONCE,
				       &previous_port);
  if (kr != KERN_SUCCESS)
    error ("notify_request failed for %x: %s.", name, mach_error_string (kr));

  if (MACH_PORT_VALID (previous_port))
    (void) mach_port_deallocate (mach_task_self (), previous_port);

  port_chain_insert (&notify_chain, name, type);
}

private void
port_chain_delete (prev_next, elem)
     port_chain_t *	prev_next;
     mach_port_t	elem;
{
  while (*prev_next && (*prev_next)->port != elem)
    prev_next = &(*prev_next)->next;

  if (*prev_next)
    *prev_next = (*prev_next)->next;    /* GCd with obstack_free() */
}

private void
port_chain_destroy (ostack)
     struct obstack *ostack;
{
  obstack_free (ostack, 0);
  obstack_init (ostack);
}

private void
port_chain_insert (list_ptr, name, type)
     port_chain_t *	list_ptr;
     mach_port_t	name;
     mach_type_t	type;
{
  kern_return_t		kr;
  port_chain_t		new;
  mach_id_t		mid;

  if (! MACH_PORT_VALID (name))
    return;
  
  if (! MACH_PORT_VALID (mid_server))
    {
      message ("Machid server port invalid, can not map port %x to MID", name);
      mid = name;
    }
  else
    {
      kr = machid_mach_register (mid_server, mid_auth, name, type, &mid);
      if (kr != KERN_SUCCESS)
	{
	  message ("Can not map port %x to MID with machid", name);
	  mid = name;
	}
    }

  new = (port_chain_t) obstack_alloc (port_chain_obstack,
				      sizeof (struct port_chain));
  new->next  = *list_ptr;
  new->port  = name;
  new->type  = type;
  new->mid   = mid;

  *list_ptr = new;
}

private port_chain_t
port_chain_lookup (list, elem)
     port_chain_t list;
     mach_port_t elem;
{
  while (list && list->port != elem)
    list = list->next;

  return list;
}

/*
 * Check that we have the given rights for the specified given port.
 */
private boolean_t
port_valid (port, mask)
  mach_port_t		port;
  mach_port_type_t	mask;
{
  kern_return_t		kr;
  mach_port_type_t	type;

  kr = mach_port_type (mach_task_self (), port, &type);
  if (kr != KERN_SUCCESS || (type & mask) != mask)
    return FALSE;

  return TRUE;
}

/*
 * Translate a PC into a symbolic address if possible.
 */
private void
print_pc (stream, pc)
     FILE *	stream;
     CORE_ADDR	pc;
{
  if (! lookup_minimal_symbol_by_pc (pc))
    fprintf_filtered (stream, local_hex_format(), pc);
  else
    {
      extern int addressprint;
      extern int asm_demangle;

      int store    = addressprint;
      addressprint = 0;
      print_address_symbolic (pc, stream, asm_demangle, "");
      addressprint = store;
    }
}

private void
print_exception (who, exception, code, sub_code)
     char *who;
     int exception, code, sub_code;
{
  int giveback = grab_terminal ();

  if (exception < 0 || exception > MAX_EXCEPTION)
    exception = 0;

  printf_filtered ("\n%s received %s exception: ", who,
		   exception_map[exception].name);

  wrap_here ("   ");

  switch(exception)
    {
      case EXC_BAD_ACCESS:
	printf_filtered ("referencing address 0x%x : %s\n", sub_code,
			 mach_error_string (code));
	break;
      case EXC_BAD_INSTRUCTION:
	printf_filtered ("illegal or undefined instruction: code %d at 0x%x\n",
			 code, sub_code);
	break;
      case EXC_ARITHMETIC:
      case EXC_EMULATION:
      case EXC_BREAKPOINT:
	printf_filtered ("code %d, sub_code %d.\n", code, sub_code);
	break;
      case EXC_SOFTWARE:
	if (code == EXC_GDB_STOPPED || code == EXC_GDB_EXITED
	    || code == EXC_GDB_SIGNALED)
	  {
	    if (code == EXC_GDB_STOPPED)
	      printf_filtered ("stopped with U*x signal %d.\n", sub_code);
	    else if (code == EXC_GDB_EXITED)
	      printf_filtered ("exited %d.\n", sub_code);
	    else
	      printf_filtered ("terminated with U*x signal %d.\n",sub_code);
	  }
	else
	  {
	    const char *ptr;

	    if (code <= 0xffff)
	      ptr = "hardware";
	    else if (code <= 0x1ffff)
	      ptr = "os emulation";
	    else
	      ptr = "user";
	    printf_filtered ("%s specific: code 0x%x, sub_code %d.\n", ptr,
			     code, sub_code);
	  }
	break;
      default:
	message ("unknown exception %d, code %d, sub_code %d.\n",
		 exception, code, sub_code);
	break;
    }

  if (giveback)
    terminal_inferior ();
}

/*
 * Try and process a message.
 */

private void
process_msg (in_msg, out_msg)
     mach_msg_header_t *in_msg;
     mach_msg_header_t *out_msg;
{
  kern_return_t	kr;

  /*
   * Check if we received a notification message or an interrupt.
   */
  if (notify_server (in_msg, out_msg) || gdb_message_server (in_msg))
    return;

  /*
   * If this message is an exception gdb_exception_server() will make an
   * upcall to gdb_exception_catch() with further action decided there.
   */
  if (! gdb_exception_server ((exception_msg_t *) in_msg,
			      (exception_reply_t *) out_msg))
    {
      message ("Ignoring unrecognized message received in process_msg: %d.",
	       in_msg->msgh_id);
      return;
    }

  /*
   * Send the exception RPC reply if somebody is expecting one and
   * we haven't forwarded the request.  If the reply port has died,
   * then deallocate the dead name.
   */
  if (MACH_PORT_VALID (out_msg->msgh_remote_port)
      && ((exception_reply_t *) out_msg)->ret_code != MIG_NO_REPLY)
    {
      kr = mach_msg_send (out_msg);
      if (kr != KERN_SUCCESS)
	{
	  message ("Unable to reply to task raising exception: %s.",
		   mach_error_string (kr));
	  if (kr == MACH_SEND_INVALID_DEST)
	    (void) mach_port_deallocate (mach_task_self (),
					 out_msg->msgh_remote_port);
	}
    }
}
      
/*
 * Handle a signal sent to the child.  This is basically exception
 * raise turned into a simple message.  We might dead-lock with ourself
 * if we wait for a reply.
 */

private void
sigchld_handler (sig)
     int	sig;
{
  kern_return_t	kr;
  WAITTYPE	status;
  int		pid;
  int		code, sub_code;

  do {
    pid = wait3(&status, WNOHANG, NULL);
    if (pid <= 0)
      {
	message ("received SIGCHLD, but wait3 returned %d.", pid);
	return;
      }
  } while (pid != inferior_pid);

  if (WIFSTOPPED(status))
    {
      code = EXC_GDB_STOPPED;
      sub_code = WSTOPSIG(status);
    }
  else if (WIFEXITED(status))
    {
      code = EXC_GDB_EXITED;
      sub_code = WEXITSTATUS(status);
    }
  else
    {
      code = EXC_GDB_SIGNALED;
      sub_code = WTERMSIG(status);
    }

  kr = gdb_exception_raise (inferior.exception_port,
			    MACH_MSG_TYPE_MAKE_SEND_ONCE, MACH_PORT_NULL, 0,
			    inferior.task, MACH_PORT_NULL,
			    EXC_SOFTWARE, code, sub_code);
  if (kr != KERN_SUCCESS)
    message ("sigchld_handler unable to send signal to main thread: %s.",
	     mach_error_string (kr));
}

/*
 * Clean up a dead thread.
 */
private void
thread_cleanup (thread)
     thread_t	thread;
{
  gdb_thread_t *gdb_thread;
  unsigned int	i;

  notify_cancel (thread);

  /*
   * This is thread_lookup(), but we need to know where in the thread
   * list to start shifting records.
   */
  for (i = 0; i < inferior.thread_count; i++)
    if (inferior.thread_list[i].port == thread)
      {
	gdb_thread = &inferior.thread_list[i];
	break;
      }

  if (i == inferior.thread_count)
    return;

  if (gdb_thread->name)
    free (gdb_thread->name);

  inferior.thread_count -= 1;
  for (/* CONTINUE */; i < inferior.thread_count; i++)
    inferior.thread_list[i] = inferior.thread_list[i + 1];

  if (thread == current_thread)
    current_thread = MACH_PORT_NULL;

  (void) mach_port_deallocate (mach_task_self (), thread);
}

private void
thread_init (thread)
     thread_t		thread;
{
  kern_return_t		kr;
  gdb_thread_t *	gdb_thread;

  gdb_thread = &inferior.thread_list[inferior.thread_count];
  inferior.thread_count += 1;

  /*
   * Most things are zero/false.
   */
  memset (gdb_thread, 0, sizeof(gdb_thread_t));

  gdb_thread->port = thread;
  notify_request (thread, MACH_TYPE_THREAD);
  gdb_thread->mid = map_port_name_to_mid (thread, MACH_TYPE_THREAD);

  kr = thread_setup (gdb_thread->mid);
  if (kr != KERN_SUCCESS)
    warning ("Newly initialized thread has disappeared: %s.",
	     mach_error_string (kr));
}

/*
 * Lookup a thread by its control port.
 */
private gdb_thread_t *
thread_lookup (thread_port)
     thread_t	thread_port;
{
  unsigned int	i;

  for (i = 0; i < inferior.thread_count; i++)
    if (inferior.thread_list[i].port == thread_port)
      return &inferior.thread_list[i];

  return NULL;
}

/*
 * Lookup a thread by its Mach id.
 */
private gdb_thread_t *
thread_lookup_by_mid (mid)
     mach_id_t mid;
{
  unsigned int	i;

  if (mid == 0)
    if (inferior.thread_count > 0)
      return &inferior.thread_list[0];
    else
      return NULL;

  for (i = 0; i < inferior.thread_count; i++)
    if (inferior.thread_list[i].mid == mid)
      return &inferior.thread_list[i];

  return NULL;
}

/*
 * Restore a thread so that the inferior can be detached.
 */
private void
thread_restore (gdb_thread_t *thread)
{
  kern_return_t	kr;

  if (thread->suspended)
    {
      kr = thread_resume(thread->port);
      if (kr != KERN_SUCCESS)
        message ("Unable to resume suspended thread %x: %s.",
		 thread->port, mach_error_string (kr));
      thread->suspended = FALSE;
    }
}

/*
 * Resume all the threads in the inferior that GDB has suspended.
 */
private void
thread_resume_all ()
{
  kern_return_t	kr;
  unsigned int	i;

  inferior_setup ();

  for (i = 0; i < inferior.thread_count; i++)
    if (inferior.thread_list[i].suspended)
      {
	inferior.thread_list[i].suspended = FALSE;
	kr = thread_resume (inferior.thread_list[i].port);
	if (kr != KERN_SUCCESS)
	  message ("Unable to resume thread %u: %s.",
		   inferior.thread_list[i].mid, mach_error_string (kr));
      }
}

/*
 * Select thread MID as the current thread and prepare it for debugging.
 * If MID is zero, then the first kernel thread is selected.
 * If MID is the task's mid and the current_thread still exists it is
 * selected, otherwise the first kernel thread is used.  This feature
 * is used by the exception mechanism.
 */
private kern_return_t
thread_setup (mid)
     mach_id_t mid;
{
  kern_return_t	kr;
  gdb_thread_t *gdb_thread;

  gdb_thread = thread_lookup_by_mid (mid);
  if (gdb_thread == NULL)
    {
      if (mid)
	message ("No thread with mid %u.", mid);
      else
	message ("No threads in inferior task %u.", inferior.mid);

      return KERN_INVALID_VALUE;
    }
  
  current_thread = gdb_thread->port;

  registers_changed ();

  if (! inferior.suspended && ! gdb_thread->suspended)
    {
      kr = thread_suspend (gdb_thread->port);
      if (kr == KERN_SUCCESS)
        {
	  gdb_thread->suspended = TRUE;
	  gdb_thread->aborted = FALSE;
	}
      else
	{
	  current_thread = MACH_PORT_NULL;
	  message ("Unable to suspend thread %u: %s.", gdb_thread->mid,
		   mach_error_string (kr));
	  return kr;
	}
    }

  if (! gdb_thread->aborted)
    {
      kr = thread_abort (gdb_thread->port);
      if (kr == KERN_SUCCESS)
	gdb_thread->aborted = TRUE;
      else
	{
	  message ("Couldn't abort system call when selecting thread %u: %s.",
		   gdb_thread->mid, mach_error_string (kr));
	  return kr;
	}
      gdb_thread->pc = read_pc ();
      gdb_thread->fp = read_register (FP_REGNUM);
      gdb_thread->sp = read_register (SP_REGNUM);

      gdb_thread->in_emulator = mach3_fetch_user_state (gdb_thread->sp,
							&gdb_thread->user_sp,
							&gdb_thread->user_fp,
							&gdb_thread->user_pc);
    }

  stop_pc = gdb_thread->pc;
  set_current_frame (create_new_frame (gdb_thread->fp, stop_pc));

  select_frame (get_current_frame (), 0);

  stop_frame_address = FRAME_FP (get_current_frame ());

  return KERN_SUCCESS;
}

/*
 * Translate a thread's state into a string.
 */
private char *
thread_state_to_string(thread)
     gdb_thread_t *	thread;
{
  static char	buf[MAX_NAME_LEN];
  gdb_thread_state_t state;
  char		swapped;
  char		kstate;
  char *	cstate;

  state = thread->state;
  swapped = state.swapped ? 'S' : ' ';

  switch (state.kernel)
    {
      case TH_STATE_RUNNING:
	kstate = 'R';
	break;
      case TH_STATE_STOPPED:
	kstate = 'S';
	break;
      case TH_STATE_WAITING:
	kstate = 'W';
	break;
      case TH_STATE_UNINTERRUPTIBLE:
	kstate = 'U';
	break;
      case TH_STATE_HALTED:
	kstate = 'H';
	break;
      case 0:
	kstate = ' ';
	break;
      default:
	kstate = '?';
	break;
    }

  if (state.has_cthread)
    switch (state.cthread)
      {
	case CPROC_RUNNING:
	  cstate = "R";
	  break;
	case CPROC_SWITCHING:
	  cstate = "S";
	  break;
	case CPROC_BLOCKED:
	  cstate = "B";
	  break;
	case CPROC_CONDWAIT:
	  cstate = "C";
	  break;
	case CPROC_CONDWAIT|CPROC_SWITCHING:
	  cstate = "CS";
	  break;
	default:
	  cstate = "?";
	  break;
      }
  else
    cstate = "";

  if (thread->suspend_count > 0)
    sprintf (buf, "%c%c%-2d   %-2.2s", kstate, swapped, thread->suspend_count,
	     cstate);
  else if (thread->suspend_count == 0)
    sprintf (buf, "%c%c     %-2.2s", kstate, swapped, cstate);
  else
    sprintf (buf, "%c%c?    %-2.2s", kstate, swapped, cstate);

  return buf;
}

/*
 * Suspend all the threads in a task.
 */
private void
thread_suspend_all ()
{
  kern_return_t	kr;
  unsigned int	i;

  inferior_setup ();

  for (i = 0; i < inferior.thread_count; i++)
    if (! inferior.thread_list[i].suspended)
      {
	inferior.thread_list[i].aborted = FALSE;

	kr = thread_suspend (inferior.thread_list[i].port);
	if (kr == KERN_SUCCESS)
	  inferior.thread_list[i].suspended = TRUE;
	else
	  message ("Unable to suspend thread %u: %s.",
		   inferior.thread_list[i].mid, mach_error_string (kr));
      }
}

/*
 * Turn on or off hardware single-stepping.
 *
 * NOTE:  We do not call thread_abort(), because the target thread is
 * either blocked in an exception or was previously aborted by thread_setup().
 * This avoids killing the exception reply port.
 */
private void
thread_trace (thread, set)
     mach_port_t thread;
     boolean_t   set;
{
  int			flavor   = TRACE_FLAVOR;
  unsigned int		stateCnt = TRACE_FLAVOR_SIZE;
  kern_return_t		kr;
  thread_state_data_t	state;
  gdb_thread_t *	gdb_thread;
  mach_id_t		mid;

#ifdef	TRACE_CLEAR
  if (! MACH_PORT_VALID (thread))
#else
  if (! MACH_PORT_VALID (thread) || ! set)
#endif
    {
      inferior.single_stepping = FALSE;
      return;
    }

  gdb_thread = thread_lookup (thread);
  if (gdb_thread)
    mid = gdb_thread->mid;
  else
    {
      message ("Unable to find thread for port %x, setting trace state anyway.",
	       thread);
      mid = 0;
    }
	       
  kr = thread_get_state(thread, flavor, state, &stateCnt);
  if (kr != KERN_SUCCESS)
    {
      message ("Unable to read state for thread %u in thread_trace: %s.",
	       mid, mach_error_string (kr));
      return;
    }
  
  inferior.single_stepping = set;

#ifdef	TRACE_CLEAR
  if (set)
    TRACE_SET (state);
  else
    TRACE_CLEAR (state);
#else
  TRACE_SET (state);
#endif

  kr = thread_set_state(thread, flavor, state, stateCnt);
  if (kr != KERN_SUCCESS)
    message ("Unable to write state for thread %u in thread_trace: %s.",
	     mid, mach_error_string (kr));
}

/*
 * Send a message to the main GDB thread to knock it out of its mach_msg wait.
 * This is used internally when the "signal" thread picks up a signal.
 * It can also be used as followsto wake up an inferior when debugging GDB.
 * 
 *	(gdb) call wakeup_main_thread ()
 *	(gdb) continue
 *
 * NOTE:  The program "gdb-stop" also implements this functionality.
 */
private void
wakeup_main_thread ()
{
  kern_return_t kr;
  mach_msg_header_t msg;

  msg.msgh_bits = MACH_MSGH_BITS(MACH_MSG_TYPE_MAKE_SEND_ONCE, 0);
  msg.msgh_size = sizeof(msg);
  msg.msgh_remote_port = gdb_message_port;
  msg.msgh_local_port  = MACH_PORT_NULL;
  msg.msgh_seqno       = 0;
  msg.msgh_id          = GDB_MESSAGE_ID_STOP;

  kr = mach_msg (&msg, MACH_SEND_MSG, sizeof(msg), 0, MACH_PORT_NULL,
		 MACH_MSG_TIMEOUT_NONE, MACH_PORT_NULL);
  if (kr != KERN_SUCCESS)
    message ("unable to send message to main thread: %s.",
	     mach_error_string (kr));
}

/*
 * Routines for handling notification messages.
 *
 * NOTE: These routines need to be 'global' because they are called
 * from the routines in libmach.
 */

kern_return_t
do_mach_notify_dead_name (notify, name)
     mach_port_t notify;
     mach_port_t name;
{
  /*
   * Find the port that died.
   */
  port_chain_t element = port_chain_lookup (notify_chain, name);

  if (element == NULL)
    {
      message ("Received dead-name notification for unknown port %x.", name);
      return KERN_SUCCESS;
    }

  switch (element->type)
    {
      case MACH_TYPE_THREAD:
	if (name == current_thread)
	  message ("\nCurrent thread %u died", element->mid);
	else
	  message ("\nThread %u died", element->mid);

	thread_cleanup (name);

	break;

      case MACH_TYPE_TASK:
	if (name != inferior.task)
	  message ("Strange: task %u died, but it was not our inferior task",
		   element->mid);
	else	       
	  {
	    message ("Inferior task %u died", element->mid);
	    inferior.task = MACH_PORT_NULL;
	  }
	break;

      default:
	message ("Dead-name notification received for %x, type = %d, mid = %u.",
		 name, element->type, element->mid);
	break;
    }

  /*
   * Delete the name from the notification list.
   */
  port_chain_delete (&notify_chain, name);

  (void) mach_port_deallocate (mach_task_self (), name);

  return KERN_SUCCESS;
}

kern_return_t
do_mach_notify_msg_accepted (notify, name)
     mach_port_t notify;
     mach_port_t name;
{
  message ("do_mach_notify_msg_accepted : notify %x, name %x", notify, name);
  return KERN_SUCCESS;
}

kern_return_t
do_mach_notify_no_senders (notify, mscount)
     mach_port_t notify;
     mach_port_mscount_t mscount;
{
  message ("do_mach_notify_no_senders : notify %x, mscount %x",
	   notify, mscount);
  return KERN_SUCCESS;
}

kern_return_t
do_mach_notify_port_deleted (notify, name)
     mach_port_t notify;
     mach_port_t name;
{
  message ("do_mach_notify_port_deleted : notify %x, name %x",
	   notify, name);
  return KERN_SUCCESS;
}

kern_return_t
do_mach_notify_port_destroyed (notify, rights)
     mach_port_t notify;
     mach_port_t rights;
{
  message ("do_mach_notify_port_destroyed : notify %x, rights %x",
	   notify, rights);
  return KERN_SUCCESS;
}

kern_return_t
do_mach_notify_send_once (notify)
     mach_port_t notify;
{
#ifdef DUMP_SYSCALL
  /* MANY of these are generated. */
  message ("do_mach_notify_send_once : notify %x", notify);
#endif
  return KERN_SUCCESS;
}

/*
 * Target interface routines.
 */

/* 
 * Start debugging the process whose unix process-id is PID.
 * A negative "pid" value is legal and signifies a mach_id not a unix pid.
 *
 * Prevent (possibly unwanted) dangerous operations by enabled users
 * like "atta 0" or "atta foo" (equal to the previous :-) and
 * "atta pidself". Anyway, the latter is allowed by specifying a MID.
 */

private void
mach3_attach (args, from_tty)
     char *args;
     int from_tty;
{
  kern_return_t kr;
  char *exec_file;
  int pid;
  mach_id_t mid;

  if (!args)
    error_no_arg ("process-id to attach");

  pid = atoi (args);

  if (pid == 0)
    error("PID=0, Debugging the master unix server does not compute.");

  /* Foo. This assumes gdb has a unix pid */
  if (pid == getpid())
    error ("I will debug myself only by MID. (Gdb would suspend itself!)");

  if (pid < 0)
      mid = -pid;
  else
      mid = 0;

#ifdef	LAST_DANGEROUS_MID
  /*
   * It can be dangerous to attach to tasks with low mids as these are
   * typically the kernel and Unix server tasks.
   */

  if (mid && mid <= LAST_DANGEROUS_MID)
    if (! query ("Are you sure you want to attach to MID %d? ", mid))
      error ("Unix server not attached.");
#endif

  if (from_tty)
    {
      exec_file = get_exec_file (0);

      if (exec_file)
	printf ("Attaching program `%s', pid %d\n", exec_file, pid);
      else
	printf ("Attaching pid %d\n", pid);

      fflush (stdout);
    }

  if (mid)
    {
      int size;

      kr = machid_mach_port (mid_server, mid_auth, mid, &inferior.task);
      if (kr != KERN_SUCCESS)
	error ("Unable to map MID %d to task port: %s", mid,
	       mach_error_string (kr));

      if (!MACH_PORT_VALID (inferior.task))
	error ("Target task died.");

      /*
       * See if the target has a U*x pid.
       */
      
      size = 0;
      if (pid_by_task (inferior.task, &pid, NULL, &size) != 0);
	pid = -mid;
    }
  else
    {
      inferior.task = task_by_pid (pid);
      if (! MACH_PORT_VALID (inferior.task))
	error ("Cannot map Unix pid %d to Mach task port", pid);
    }

  kr = task_suspend (inferior.task);
  if (kr != KERN_SUCCESS)
    {
      (void) mach_port_deallocate (mach_task_self(), inferior.task);
      inferior.task = MACH_PORT_NULL;
      error ("Cannot suspend inferior: task died.");
    }

  inferior.suspended = TRUE;
  inferior.setup = FALSE;

  attach_flag = TRUE;
  inferior_pid = pid;

  /*
   * Intialize the inferior state for debugging.
   */
  push_target (&mach3_ops);

  inferior_init ();

  inferior_setup ();

  /*
   * Rely on thread_setup to print any message that is appropriate.
   * As long as the task exists, we want to continue.  The user may
   * want to figure out why there are no threads in the task.
   */
  (void) thread_setup (0);

#ifdef	PT_ATTACH
  if (our_ptrace (PT_ATTACH, 0) != KERN_SUCCESS)
#endif
  warning ("GDB unable to enable U*x tracing.");
}

/*
 * If we are closed off for some reason, clean up our data structures.
 * This will usually be a no-op as it is normally called after we have
 * already cleaned up.
 */
private void
mach3_close(quitting)
     int quitting;
{
  if (MACH_PORT_VALID (inferior.task))
    mach3_mourn_inferior ();
}

/*
 * Stop debugging our inferior and detach it.  The program resumes
 * execution and will no longer stop on signals, etc.  We'd better
 * not have left any breakpoints in the program or it'll die when it
 * hits one.
 *
 * If an argument is present is the SIGNAL to pass to the inferior
 * when it is resumed.
 */

private void
mach3_detach (args, from_tty)
     char *args;
     int from_tty;
{
  int send_signal = 0;

  inferior_check ();

  if (from_tty)
    {
      char *exec_file = get_exec_file (0);
      if (exec_file == 0)
	exec_file = "";
      printf ("Detaching program: %s pid %d\n", exec_file, inferior_pid);
      fflush (stdout);
    }

  if (args)
    send_signal = atoi (args);
  
  inferior_restore (send_signal);
  inferior_resume (send_signal);
  mach3_mourn_inferior ();
}

/*
 * Print status information about what we're accessing.
 */
private void
mach3_files_info (ignore)
     struct target_ops *ignore;
{
  printf ("\tUsing the running image of %s process %d.\n",
	  attach_flag? "attached": "child", inferior_pid);
}

private void
mach3_kill ()
{
  if (MACH_PORT_VALID (inferior.task))
    {
      if (inferior.has_ptrace_server)
        {
	  /*
	   * Make sure the server thinks the inferior is stopped before
	   * killing it.
	   */
	  if (! inferior.stopped_in_server)
	    kill (inferior_pid, SIGSTOP);
	  if (our_ptrace (PT_KILL, SIGKILL) != KERN_SUCCESS)
	    message ("U*x kill failed.");
	}

      /*
       * One more nail in the coffin.
       */
      (void) task_terminate (inferior.task);

      /*
       * If this was a U*x process collect its status.
       */
      if (inferior_pid > 0)
        {
	  WAITTYPE dummy;

	  wait3 (&dummy, WNOHANG, NULL);
	}
    }

  mach3_mourn_inferior ();
}

/* ARGSUSED */
private void
mach3_open (arg, from_tty)
     char *arg;
     int from_tty;
{
  error ("Use the \"run\" command to start a Unix child process.");
}

/*
 * Get ready to modify the registers array.
 */
private void
mach3_prepare_to_store ()
{
  read_register_bytes (0, NULL, REGISTER_BYTES);
}

/*
 * Resume execution of the inferior process.
 * If STEP is nonzero, then single-step it.
 * If SIGNAL is nonzero, then give it that signal.
 */
private void
mach3_resume (step, send_signal)
     int step;
     int send_signal;
{
  kern_return_t	kr;
  gdb_thread_t *gdb_thread;

  if (step)
    {
      thread_basic_info_data_t	th_info;
      mach_msg_type_number_t	infoCnt = THREAD_BASIC_INFO_COUNT;

      if (inferior.exc_list && inferior.exc_list->next)
	{
	  message ("There are still exceptions outstanding.  GDB will become");
	  message ("confused if you do not discard them before stepping.");
	  if (! query ("Do you want to discard them? "))
	    {
	      message ("Use 'task next-exception' to examine the exceptions.");
	      error ("Thread not stepped.");
	    }

	  while (inferior.exc_list)
	    inferior_exception_discard ();
	}

      /*
       * There is no point in single stepping when current_thread is dead.
       */
      if (! MACH_PORT_VALID (current_thread))
	error ("No thread selected; cannot single step");
      
      /*
       * If current_thread is suspended, tracing it would never return.
       */
      kr = thread_info (current_thread, THREAD_BASIC_INFO,
			(thread_info_t) &th_info, &infoCnt);
      if (kr != KERN_SUCCESS)
	{
	  message ("mach3_resume: cannot get thread info: %s",
		   mach_error_string (kr));

	  inferior_check ();

	  error ("Did the thread die?");
	}
      
      if (th_info.suspend_count)
	error ("Can't trace a suspended thread. Use \"thread resume\" command to resume it.");

      gdb_thread = thread_lookup (current_thread);
      if (gdb_thread == NULL)
	error ("Unable to find the current thread record.");

      thread_suspend_all ();

      thread_trace (current_thread, TRUE);
      
      kr = thread_resume (current_thread);
      if (kr == KERN_SUCCESS)
	gdb_thread->suspended = FALSE;
      else
	{
	  if (kr == KERN_FAILURE)
	    {
	      gdb_thread->suspended = FALSE;
	      message ("Somebody else is playing with this task.");
	      message ("The thread I just suspended is no longer suspended!");
	      if (! query ("Do you want me to keep going? "))
		  error ("Thread not stepped.");
	    }
	  else if (kr == KERN_INVALID_ARGUMENT)
	    {
	      current_thread = MACH_PORT_NULL;
	      error ("The current thread died.");
	    }
	  else
	    error ("mach3_resume: thread_resume strange error: %s",
		   mach_error_string (kr));
	}
    }
  else
    thread_resume_all ();

  inferior_resume (send_signal);
}

/*
 * Wait for child to do something.  Return pid of child, or -1 in case
 * of error.  Child's exit status returned through our argument.
 */
private int
mach3_wait (status)
     WAITTYPE *status;
{
  kern_return_t	kr;
  exc_rec_t *	exc_rec;
  union msg {	/* Change this if you add bigger GDB-internal messages. */
    mach_msg_header_t	header;
    exception_msg_t	exception_msg;
  } in_msg, out_msg;

  if (! MACH_PORT_VALID (inferior.task))
    return -1;

  /*
   * If we already have exceptions queued just handle the next one.
   */
  if (inferior.exc_list == NULL)
    for (;;)
      {
	/*
	 * Either notify (death), exception or message can stop the inferior.
	 */
	mutex_lock (&gdb_quit_lock);
	/*
	 * This is the QUIT-macro with the addition of the mutex_unlock call.
	 */
	if (quit_flag)
	  {
	    mutex_unlock (&gdb_quit_lock);
	    break;
	  }
	gdb_really_waiting = TRUE;
	mutex_unlock (&gdb_quit_lock);

	kr = mach_msg (&in_msg.header,		/* header */
		       MACH_RCV_MSG,		/* options */
		       0,				/* send size */
		       sizeof (union msg),	/* receive size */
		       gdb_port_set,		/* receive name */
		       MACH_MSG_TIMEOUT_NONE,
		       MACH_PORT_NULL);

	mutex_lock (&gdb_quit_lock);
	gdb_really_waiting = FALSE;
	mutex_unlock (&gdb_quit_lock);

	if (kr == KERN_SUCCESS)
	  {
	    process_msg (&in_msg.header, &out_msg.header);
	    /*
	     * If inferior.task is NULL then the inferior has died, and
	     * we should return to the command level.
	     */
	    if (inferior.task == MACH_PORT_NULL || inferior.exc_list)
	      break;
	  }
	else if (kr == MACH_RCV_TOO_LARGE)
	  message ("Unable to receive over-sized message in mach3_wait.");
	else
	  {
	    message ("mach_msg (receive) failed: %s.", mach_error_string (kr));
	    quit_flag = 1;
	    break;
	  }
      }

  kr = inferior_suspend ();
  if (kr != KERN_SUCCESS)
    message ("Found task dead in mach3_wait.");
  
  /*
   * Process any waiting messages.
   */
  do {
    kr = mach_msg (&in_msg.header, MACH_RCV_MSG|MACH_RCV_TIMEOUT, 0,
		   sizeof (union msg), gdb_port_set, 0, MACH_PORT_NULL);
    if (kr == KERN_SUCCESS)
      process_msg (&in_msg.header, &out_msg.header);
    else if (kr == MACH_RCV_TOO_LARGE)
      {
	message ("Unable to receive over-sized message while cleaning up.");
	kr = KERN_SUCCESS;
      }
    else if (kr != MACH_RCV_TIMED_OUT)
      {
	message ("mach_msg failed while cleaning up: %s.",
		 mach_error_string(kr));
	quit_flag = 1;
      }
  } while (kr == KERN_SUCCESS);

  if (MACH_PORT_VALID (inferior.task))
    inferior_setup ();

  exc_rec = inferior_exception_next ();

  QUIT;

  /*
   * Set the U*x status to an appropriate value.
   */
  if (exc_rec)
    {
      if (exc_rec->exception == EXC_SOFTWARE
	  && (exc_rec->code == EXC_GDB_STOPPED
	      || exc_rec->code == EXC_GDB_EXITED
	      || exc_rec->code == EXC_GDB_SIGNALED))
	{
	  if (exc_rec->code == EXC_GDB_STOPPED)
	    WSETSTOP (*status, exc_rec->sub_code);
	  else if (exc_rec->code == EXC_GDB_EXITED)
	    WSETEXIT (*status, exc_rec->sub_code);
	  else
	    *status = exc_rec->sub_code;
	}
      else if (exception_map[exc_rec->exception].sigmap != SIG_UNKNOWN)
	WSETSTOP (*status, exception_map[exc_rec->exception].sigmap);
      else
	{
	  /*
	   * At present, this is only EXC_SOFTWARE.  It may be possible
	   * to map some codes in this class to signals based on
	   * ux_exception.h.  However, I don't know whether these
	   * mappings depend on the particular U*x server you have.
	   * For now just report that the task stopped and print
	   * the exception if the user hasn't seen it before.
	   */
	  WSETSTOP (*status, SIG_UNKNOWN);
	  if (! exception_map[exc_rec->exception].print)
	    print_exception ("Thread", exc_rec->exception, exc_rec->code,
			     exc_rec->sub_code);
	}

      if (exc_rec->next)
        message ("There are multiple exceptions pending.  Handling first.");
    }
  else if (! MACH_PORT_VALID (inferior.task))
    WSETEXIT (*status, 0);
  else
    error ("Strange, we stopped without an exception and the task is alive.");

  return inferior_pid;
}

/*
 * Return 0 on failure, number of bytes handled otherwise.
 */
private int
mach3_xfer_memory (memaddr, myaddr, len, write, target)
     CORE_ADDR memaddr;
     char *myaddr;
     int len;
     int write;
     struct target_ops *target;	/* IGNORED */
{
  int result;

  if (write)
    result = inferior_write (memaddr, myaddr, len);
  else
    result = inferior_read  (memaddr, myaddr, len);

  return result;
}

/*
 * Start an inferior Unix child process.
 *
 * EXEC_FILE is the file to run.
 * ALLARGS is a string containing the arguments to the program.
 * ENV is the environment vector to pass.
 *
 * NOTE: It would be nice if we were also be able to start pure Mach 3.0
 * tasks.  To do that would require a generic loader with a small amount
 * of target dependent code.
 */

private void
mach3_create_inferior (exec_file, allargs, env)
      char *exec_file;
      char *allargs;
      char **env;
{
  if (mach3_fork_inferior_func == NULL)
    error ("You can't do that with this version of GDB, try 'attach'.");

  /*
   * The inferior will have the same ptrace facilities as GDB.
   */
  inferior.has_ptrace_server = gdb_has_ptrace_server;

  (*mach3_fork_inferior_func) (exec_file, allargs, env, inferior_setup_self,
			       inferior_prepare);
  /*
   * We are at the first instruction we care about.  Go for it...
   */
  proceed ((CORE_ADDR) -1, 0, 0);
}

/*
 * Clean up the inferior task.  It has either died or been detached.
 */
private void
mach3_mourn_inferior ()
{
  kern_return_t kr;

  /*
   * Cancel any outstanding notification requests.
   * Notify_cancel will delete the record.
   */
  while (notify_chain)
    notify_cancel (notify_chain->port);

  port_chain_destroy (port_chain_obstack);

  if (MACH_PORT_VALID (inferior.notify_port))
    {
      kr = mach_port_mod_refs (mach_task_self (), inferior.notify_port,
			       MACH_PORT_RIGHT_RECEIVE, -1);
      if (kr != KERN_SUCCESS)
	message ("Could not deallocate our notify port: %s.",
		 mach_error_string (kr));
    }
  inferior.notify_port = MACH_PORT_NULL;

  while (inferior.exc_list)
    inferior_exception_discard ();

  if (MACH_PORT_VALID (inferior.exception_port))
    {
      kr = mach_port_mod_refs (mach_task_self (), inferior.exception_port,
			       MACH_PORT_RIGHT_RECEIVE, -1);
      if (kr != KERN_SUCCESS)
	message ("Could not deallocate our exception port: %s.",
		 mach_error_string (kr));
    }
  inferior.exception_port = MACH_PORT_NULL;

  if (MACH_PORT_VALID (inferior.old_exception_port))
    {
      kr = mach_port_deallocate (mach_task_self(), inferior.old_exception_port);
      if (kr != KERN_SUCCESS)
	message ("Could not deallocate inferior's old exception port: %s.",
		 mach_error_string (kr));
    }
  inferior.old_exception_port = MACH_PORT_NULL;

  /*
   * Clean up any remaining threads.
   */
  while (inferior.thread_count)
    thread_cleanup (inferior.thread_list[inferior.thread_count - 1].port);
  free (inferior.thread_list);

  while (inferior.cthread_count--)
    if (inferior.cthread_list[inferior.cthread_count].name)
      free (inferior.cthread_list[inferior.cthread_count].name);
  free (inferior.cthread_list);

  if (current_thread != MACH_PORT_NULL)
    {
      message ("Strange all threads cleaned up, but not the current thread %x.",
	       current_thread);
      (void) mach_port_deallocate (mach_task_self (), current_thread);
      current_thread = MACH_PORT_NULL;
    }

  if (MACH_PORT_VALID (inferior.task))
    {
      kr = mach_port_deallocate (mach_task_self(), inferior.task);
      if (kr != KERN_SUCCESS)
	message ("Could not deallocate inferior's task port: %s.",
		 mach_error_string (kr));
    }
  inferior.task = MACH_PORT_NULL;

  inferior.mid = 0;
  inferior.emulator_present = FALSE;
  inferior.single_stepping = FALSE;
  inferior.suspended = FALSE;
  inferior.setup = FALSE;
  inferior.stopped_in_server = FALSE;
  inferior.has_ptrace_server = gdb_has_ptrace_server;
  inferior.thread_count = 0;
  inferior.thread_list = NULL;
  inferior.cthread_count = 0;
  inferior.cthread_list = NULL;
  
  attach_flag = 0;
  inferior_pid = 0;

  unpush_target (&mach3_ops);		/* Pop out of handling an inferior */

  generic_mourn_inferior ();
}

private int
mach3_can_run ()
{
  return TRUE;
}

/*
 * Mach 3.0 commands.
 */

/*
 * Exception Commands.
 */

/*
 * You may either FORWARD the exception to the inferior, or KEEP
 * it and return to GDB command level.
 *
 * exception mid [ forward | keep ]
 */

private void
exception_command (args, from_tty)
     char *args;
     int from_tty;
{
  char *scan = args;
  int exception;

  if (!args)
    error_no_arg ("exception number action");

  while (*scan == ' ' || *scan == '\t') scan++;
  
  if (isdigit(*scan))
    while (isdigit(*scan))
      scan++;
  else
    error ("exception number action");

  exception = atoi (args);
  if (exception <= 0 || exception > MAX_EXCEPTION)
    error ("Allowed exception numbers are in range 1..%d", MAX_EXCEPTION);

  if (isspace(*scan))
    while (isspace(*scan)) scan++;
  else
    error ("exception number must be followed by a space");

  args = scan;
  while (*scan)
    scan++;

  if (scan == args)
    error("exception number action");

  if (strncmp (args, "forward", scan - args) == 0)
    exception_map[ exception ].forward = TRUE;
  else if (strncmp (args, "keep", scan - args) == 0)
    exception_map[ exception ].forward = FALSE;
  else
    error ("exception action is either \"keep\" or \"forward\"");
}

private void
print_exception_info (exception)
     int exception;
{
  boolean_t forward = exception_map[ exception ].forward;

  printf_filtered ("%s\t(%d): ", exception_map[ exception ].name,
		   exception);
  if (!forward)
    if (exception_map[ exception ].sigmap != SIG_UNKNOWN)
      printf_filtered ("keep and handle as signal %d\n",
		       exception_map[ exception ].sigmap);
    else
      printf_filtered ("keep and handle as unknown signal %d\n",
		       exception_map[ exception ].sigmap);
  else
    printf_filtered ("forward exception to inferior\n");
}

private void
exception_info (args, from_tty)
     char *args;
     int from_tty;
{
  int exception;

  if (!args)
    for (exception = 1; exception <= MAX_EXCEPTION; exception++)
      print_exception_info (exception);
  else
    {
      exception = atoi (args);

      if (exception <= 0 || exception > MAX_EXCEPTION)
	error ("Invalid exception number, values from 1 to %d allowed",
	       MAX_EXCEPTION);
      print_exception_info (exception);
    }
}

/*
 * Commands for setting and displaying the name of GDB's message port.
 */

/*
 * Register our message port with the name server.
 *
 * Currently used only by the external gdb-stop program.
 *
 * NOTE: If the message port is also used for other purposes, then
 * this registration can be a security hole.
 */

private void
port_name_set_command (args, from_tty)
     char *args;
     int from_tty;
{
  kern_return_t kr;
  size_t len;

  if (args == NULL || *args == '\0')
    error ("Invalid registration request: missing or empty name.");

  len = strlen (args);
  if (len > (MAX_NAME_LEN - sizeof(GDB_NAME_PREFIX)))
    error ("Invalid registration request: name too long: %u > %u.",
	   len + sizeof(GDB_NAME_PREFIX), MAX_NAME_LEN);

  if (gdb_name_registered)
    {
      kr = netname_check_out (name_server_port, gdb_port_name,
			      mach_task_self ());
      if (kr != KERN_SUCCESS)
	message ("Unable to un-register previous name '%s': %s.",
		 gdb_port_name, mach_error_string (kr));

      gdb_name_registered = FALSE;
    }

  memcpy (&gdb_port_name[sizeof(GDB_NAME_PREFIX) - 1], args, len + 1);

  kr = netname_check_in (name_server_port, gdb_port_name, mach_task_self (),
			 gdb_message_port);
  if (kr != KERN_SUCCESS)
    error ("Unable to register new name '%s': %s.", gdb_port_name,
	   mach_error_string (kr));

  gdb_name_registered = TRUE;

  if (from_tty)
    message ("GDB's message port name is now '%s'.", gdb_port_name);
}

private void
port_name_show_command (args, from_tty)
     char *args;
     int from_tty;
{
  if (gdb_name_registered)
    message ("GDB's message port name is '%s'.", gdb_port_name);
  else
    message ("GDB's message port is not currently registered.");
}

/*
 * Task specific commands.
 */

/*ARGSUSED*/
private void
task_command (arg, from_tty)
     char *arg;
     int from_tty;
{
  printf ("\"task\" must be followed by the name of a task command.\n");
  help_list (cmd_task_list, "task ", -1, stdout);
}


private void
task_next_exception_command (args, from_tty)
     char *args;
     int from_tty;
{
  exc_rec_t *	exc_rec;

  inferior_check ();

  if (args)
    message ("Ignoring arguments");

  inferior_exception_discard ();
  exc_rec = inferior_exception_next ();
  if (exc_rec == NULL)
    error ("There are no more exceptions.");

  print_exception (exc_rec->mid != inferior.mid ? "Thread" : "Task",
		   exc_rec->exception, exc_rec->code, exc_rec->sub_code);
}

private void
task_resume_command (args, from_tty)
     char *args;
     int from_tty;
{
  kern_return_t			kr;
  mach_msg_type_number_t	infoCnt = TASK_BASIC_INFO_COUNT;
  task_basic_info_data_t	ta_info;
  
  inferior_check ();

  if (args)
    message ("Ignoring arguments.");

  kr = task_info (inferior.task, TASK_BASIC_INFO, (task_info_t) &ta_info,
		  &infoCnt);
  if (kr != KERN_SUCCESS)
    error ("task_resume_command: task_info: %s", mach_error_string (kr));
  
  if (ta_info.suspend_count == 0)
    error ("Inferior task %u is not suspended", inferior.mid);
  else if (ta_info.suspend_count == 1 && from_tty &&
	   ! query ("Suspend count is now 1. Do you know what you are doing? "))
    error ("Task not resumed");

  kr = task_resume (inferior.task);
  if (kr != KERN_SUCCESS)
    {
      if (kr == KERN_FAILURE)
	{
	  /*
	   * This will cause us to clean up below.
	   */
	  ta_info.suspend_count = 1;
	}
      else if (kr == KERN_INVALID_ARGUMENT)
        {
	  mach3_mourn_inferior ();
	  error ("Inferior task died.");
	}
      else
	error ("task_resume_command: task_resume: %s", mach_error_string (kr));
    }

  if (ta_info.suspend_count == 1)
    {
      message ("Inferior task %u is no longer suspended", inferior.mid);

      registers_changed ();

      inferior.suspended = FALSE;
    }
  else
    message ("Inferior task %u suspend count is now %d",
	     inferior.mid, ta_info.suspend_count-1);
}

private void
task_suspend_command (args, from_tty)
     char *args;
     int from_tty;
{
  kern_return_t			kr;
  boolean_t			was_suspended;
  mach_msg_type_number_t	infoCnt = TASK_BASIC_INFO_COUNT;
  task_basic_info_data_t	ta_info;
  
  inferior_check ();

  if (args)
    message ("Ignoring arguments.");

  was_suspended = inferior.suspended;
  kr = inferior_suspend ();
  if (kr != KERN_SUCCESS)
    {
      mach3_mourn_inferior ();
      message ("Inferior task died.");
    }

  kr = task_info (inferior.task, TASK_BASIC_INFO, (task_info_t) &ta_info,
		  &infoCnt);
  if (kr != KERN_SUCCESS)
    error ("task_suspend_command: task_info: %s", mach_error_string (kr));
  
  if (was_suspended && ta_info.suspend_count == 1)
    {
      unsigned int i;

      message ("This is strange.  GDB had thought the inferior was suspended.");
      /*
       * Try to recover gracefully.  (Could do more.)
       */
      was_suspended = FALSE;
      for (i = 0; i < inferior.thread_count; i++)
        if (! inferior.thread_list[i].suspended)
	  inferior.thread_list[i].aborted = FALSE;
    }

  inferior.suspended = was_suspended;

  message ("Inferior task %u suspend count is now %d.",
	   inferior.mid, ta_info.suspend_count);

  if (! was_suspended && query ("Set up the task for debugging? "))
    {
      inferior.suspended = TRUE;
      inferior.setup = FALSE;
      inferior_setup ();
      thread_setup (0);
    }
}

private char *
get_size (bytes)
     int bytes;
{
  static char size [ 30 ];
  int zz = bytes/1024;

  if (zz / 1024)
    sprintf (size, "%-2.1f M", ((float)bytes)/(1024.0*1024.0));
  else
    sprintf (size, "%d K", zz);

  return size;
}

/* Does this require the target task to be suspended?? I don't think so. */
private void
task_info_command (args, from_tty)
     char *	args;
     int	from_tty;
{
  task_t			task;
  kern_return_t			kr;
  mach_id_t			mid = 0;
  thread_array_t		thread_list;
  mach_msg_type_number_t	thread_count;
  task_basic_info_data_t	ta_info;
  mach_msg_type_number_t	infoCnt = TASK_BASIC_INFO_COUNT;
  
  if (MACH_PORT_VALID (inferior.task))
    mid = map_port_name_to_mid (inferior.task, MACH_TYPE_TASK);

  task = inferior.task;

  if (args)
    {
      int tmid = atoi (args);

      if (tmid <= 0)
	error ("Invalid mid %d for task info.", tmid);

      if (tmid != mid)
	{
	  mid = tmid;
	  kr = machid_mach_port (mid_server, mid_auth, mid, &task);
	  if (kr != KERN_SUCCESS)
	    error ("Unable to map MID to port: %s.", mach_error_string (kr));
	}
    }

  if (mid == 0)
    error ("You have to give the task MID as an argument.");

  kr = task_info (task, TASK_BASIC_INFO, (task_info_t) &ta_info, &infoCnt);
  if (kr != KERN_SUCCESS)
    {
      if (task != inferior.task)
	(void) mach_port_deallocate (mach_task_self (), task);
      error ("task_info_command: task_info: %s", mach_error_string (kr));
    }

  printf_filtered ("\nTask info for task %u:\n\n", mid);
  printf_filtered (" Suspend count : %d\n", ta_info.suspend_count);
  printf_filtered (" Base priority : %d\n", ta_info.base_priority);
  printf_filtered (" Virtual size  : %s\n", get_size (ta_info.virtual_size));
  printf_filtered (" Resident size : %s\n", get_size (ta_info.resident_size));

  kr = task_threads (task, &thread_list, &thread_count);
  if (kr != KERN_SUCCESS)
    {
      if (task != inferior.task)
	(void) mach_port_deallocate (mach_task_self (), task);
      if (kr == KERN_INVALID_ARGUMENT)
	error ("Target task died.");
      else
	error ("task_info_command: task_threads: %s", mach_error_string (kr));
    }

  printf_filtered (" Thread count  : %d\n", thread_count);

  consume_thread_list (thread_list, thread_count);

  if (have_emulator_p (task))
    printf_filtered (" Emulator at   : 0x%x..0x%x\n",
		     EMULATOR_BASE, EMULATOR_END);
  else
    printf_filtered (" No emulator.\n");

  if (task != inferior.task)
    (void) mach_port_deallocate (mach_task_self (), task);
  else if (thread_count)
    printf_filtered ("\nUse the \"thread list\" command to see the threads\n");
}


/*
 * Thread specific commands.
 */

/*ARGSUSED*/
private void
thread_command (arg, from_tty)
     char *	arg;
     int	from_tty;
{
  printf ("\"thread\" must be followed by the name of a thread command.\n");
  help_list (cmd_thread_list, "thread ", -1, stdout);
}

private void
thread_kill_command (args, from_tty)
     char *args;
     int from_tty;
{
  mach_id_t	mid;
  gdb_thread_t *gdb_thread;
  kern_return_t kr;

  inferior_check ();

  mid = mach3_parse_thread_id (args);

  if (mid == 0)
    {
      if (from_tty && !query("Do you really want to kill the current thread? "))
	error ("Thread not killed.");
      if (! MACH_PORT_VALID (current_thread))
	error ("No current thread.");

      gdb_thread = thread_lookup (current_thread);
    }
  else
    gdb_thread = thread_lookup_by_mid (mid);

  if (gdb_thread == NULL)
    error ("No thread with MID %u.", mid);

  kr = thread_terminate (gdb_thread->port);
  if (kr != KERN_SUCCESS)
    error ("Thread %u could not be terminated: %s.", gdb_thread->mid,
	   mach_error_string (kr));
  else
    message ("Thread %u killed.", gdb_thread->mid);

  if (gdb_thread->port == current_thread)
    current_thread = MACH_PORT_NULL;

  thread_cleanup (gdb_thread->port);

  if (current_thread == MACH_PORT_NULL)
    (void) thread_setup (0);
}

/*
 * List all the kernel and C threads in the inferior.
 */
private void
thread_list_command(arg, from_tty)
     char *	arg;
     int	from_tty;
{
  gdb_thread_t *		gdb_thread;
  kern_return_t			kr;
  char *			fmt;
  unsigned int			i;
  mach_msg_type_number_t	infoCnt;
  thread_basic_info_data_t	ths;
  
  inferior_check ();
  
  if (inferior.thread_count == 1)
    fmt = "There is %d kernel thread in task %d.\n";
  else
    fmt = "There are %d kernel threads in task %d.\n";
  
  printf_filtered (fmt, inferior.thread_count, inferior.mid);
  
  puts_filtered (TL_HEADER);
  
  inferior_setup ();

  for (i = 0; i < inferior.thread_count; i++)
    {
      gdb_thread = &inferior.thread_list[i];
      
      infoCnt = THREAD_BASIC_INFO_COUNT;

      kr = thread_info (gdb_thread->port, THREAD_BASIC_INFO,
			(thread_info_t) &ths, &infoCnt);
      if (kr == KERN_SUCCESS)
	{
	  gdb_thread->state.kernel = ths.run_state;
	  gdb_thread->state.swapped = (ths.flags & TH_FLAGS_SWAPPED);
	  gdb_thread->suspend_count = ths.suspend_count;
	}
      else
	{
	  message ("Unable to get basic info on thread %u : %s",
		   gdb_thread->mid, mach_error_string (kr));
	  gdb_thread->state.kernel = 0;
	  gdb_thread->state.swapped = 0;
	  gdb_thread->suspend_count = -1;
	}

      printf_filtered (TL_K_FORMAT, i, gdb_thread->mid,
		       gdb_thread->port == current_thread ? '*' : ' ',
		       gdb_thread->name,
		       gdb_thread->in_emulator ? 'E' : ' ',
		       thread_state_to_string (gdb_thread),
		       gdb_thread->wired ? "wired" : "");
      print_pc (stdout, gdb_thread->pc);
      puts_filtered ("\n");
    }

  for (i = 0; i < inferior.cthread_count; i++)
    {
      gdb_thread = &inferior.cthread_list[i];

      printf_filtered (TL_C_FORMAT, i + 1, gdb_thread->name,
		       thread_state_to_string (gdb_thread));
      print_pc (stdout, gdb_thread->pc);
      puts_filtered ("\n");
    }
}

private void
thread_resume_command (args, from_tty)
     char *	args;
     int	from_tty;
{
  kern_return_t			kr;
  gdb_thread_t *		gdb_thread;
  mach_id_t			mid;
  mach_msg_type_number_t	infoCnt = THREAD_BASIC_INFO_COUNT;
  thread_basic_info_data_t	th_info;
  
  inferior_check ();

  if (! strcmp (args, "all")) {
    thread_resume_all ();
    return;
  }

  mid = mach3_parse_thread_id (args);

  if (mid == 0)
    {
      if (! MACH_PORT_VALID (current_thread))
	error ("No current thread.");

      gdb_thread = thread_lookup (current_thread);
    }
  else
      gdb_thread = thread_lookup_by_mid (mid);

  if (gdb_thread == NULL)
    error ("No thread with MID %u.", mid);

  if (! gdb_thread->suspended
      && !query("GDB doesn't think thread %u is suspended.  Resume it anyway? ",
		 gdb_thread->mid))
    error ("Thread %u not resumed.", gdb_thread->mid);
       
  kr = thread_info (gdb_thread->port, THREAD_BASIC_INFO,
		    (thread_info_t) &th_info, &infoCnt);
  if (kr != KERN_SUCCESS)
    error ("Can't get info for thread %u: %s.", gdb_thread->mid,
	   mach_error_string (kr));
  
  if (th_info.suspend_count > 0)
    {
      kr = thread_resume (gdb_thread->port);
      if (kr == KERN_SUCCESS)
	{
	  message ("Thread %u suspend count is %d.", gdb_thread->mid,
		   th_info.suspend_count-1);
	  if (th_info.suspend_count == 1)
	    gdb_thread->suspended = FALSE;
	  else if (query ("Do you want GDB to consider the thread suspended?"))
	    {
	      if (! gdb_thread->suspended)
		gdb_thread->aborted = FALSE;
	      gdb_thread->suspended = TRUE;
	    }
	}
      else
	message ("Unable to resume thread %u: %s", gdb_thread->mid,
		 mach_error_string (kr));
    }
  else
    message ("Thread %u is not suspended.", gdb_thread->mid);
}

private void
thread_select_command (args, from_tty)
     char *args;
     int from_tty;
{
  mach_id_t mid;

  inferior_check ();

  mid = mach3_parse_thread_id (args);

  if (mid == 0)
    error ("You must select threads by MID or @SLOTNUMBER");

  if (thread_setup (mid) != KERN_SUCCESS)
    return;

  if (from_tty)
    printf_filtered ("Thread %u selected.\n", mid);
}

private void
thread_suspend_command (args, from_tty)
     char *args;
     int from_tty;
{
  kern_return_t			kr;
  gdb_thread_t *		gdb_thread;
  mach_id_t			mid;
  mach_msg_type_number_t	infoCnt;
  thread_basic_info_data_t	th_info;
  
  inferior_check ();

  if (! strcmp (args, "all")) {
    thread_suspend_all ();
    return;
  }

  mid = mach3_parse_thread_id (args);

  if (mid == 0)
    {
      if (! MACH_PORT_VALID (current_thread))
	error ("No current thread.");

      gdb_thread = thread_lookup (current_thread);
    }
  else
    gdb_thread = thread_lookup_by_mid (mid);

  if (gdb_thread == NULL)
    error ("No thread with MID %u.", mid);

  if (gdb_thread->suspended
      && ! query ("GDB thinks thread %u is suspended.  Suspend it again? ",
		  gdb_thread->mid))
    error ("Thread %u not suspended.", gdb_thread->mid);

  kr = thread_suspend (gdb_thread->port);
  if (kr != KERN_SUCCESS)
    error ("Unable to suspend thread %u: %s.", gdb_thread->mid,
	   mach_error_string (kr));

  infoCnt = THREAD_BASIC_INFO_COUNT;
  kr = thread_info (gdb_thread->port, THREAD_BASIC_INFO,
		     (thread_info_t) &th_info, &infoCnt);
  if (kr == KERN_SUCCESS)
    {
      message ("Thread %u suspend count is %d.", gdb_thread->mid,
	       th_info.suspend_count);
      if (! gdb_thread->suspended)
	gdb_thread->aborted = FALSE;
      gdb_thread->suspended = TRUE;
    }
  else
    message ("Unable to get thread_info for %u: %s.",
	     gdb_thread->mid, mach_error_string (kr));
}

private void
add_mach_specific_commands ()
{
  extern void condition_thread ();

  /* Thread handling commands */

  add_prefix_cmd ("thread", class_stack, thread_command,
		  "Generic command for handling threads in the debugged task.",
		  &cmd_thread_list, "thread ", 0, &cmdlist);

  add_com_alias ("th", "thread", class_stack, 1);

  add_cmd ("select", class_stack, thread_select_command, 
	   "Select and print MID of the selected thread",
	   &cmd_thread_list);
  add_cmd ("list",   class_stack, thread_list_command,
	   "List info of task's threads. Selected thread is marked with '*'",
	   &cmd_thread_list);
  add_cmd ("suspend", class_run, thread_suspend_command,
	   "Suspend one or all of the threads in the selected task.",
	   &cmd_thread_list);
  add_cmd ("resume", class_run, thread_resume_command,
	   "Resume one or all of the threads in the selected task.",
	   &cmd_thread_list);
  add_cmd ("kill", class_run, thread_kill_command,
	   "Kill the specified thread MID from inferior task.",
	   &cmd_thread_list);
  add_cmd ("break", class_breakpoint, condition_thread,
	   "Breakpoint N will only be effective for thread MID or @SLOT\n\
	    If MID/@SLOT is omitted allow all threads to break at breakpoint",
	   &cmd_thread_list);

  /* Thread command shorthands (for backward compatibility) */
  add_alias_cmd ("ts", "thread select", 0, 0, &cmdlist);
  add_alias_cmd ("tl", "thread list",   0, 0, &cmdlist);

  /* task handling commands */

  add_prefix_cmd ("task", class_stack, task_command,
		  "Generic command for handling debugged task.",
		  &cmd_task_list, "task ", 0, &cmdlist);

  add_com_alias ("ta", "task", class_stack, 1);

  add_cmd ("next-exception", class_run, task_next_exception_command,
	   "Set up GDB to examine the next pending exception.",
	   &cmd_task_list);
  add_cmd ("suspend", class_run, task_suspend_command,
	   "Suspend the inferior task.",
	   &cmd_task_list);
  add_cmd ("resume", class_run, task_resume_command,
	   "Resume the inferior task.",
	   &cmd_task_list);
  add_cmd ("info", no_class, task_info_command,
	   "Print information about the specified task.",
	   &cmd_task_list);

  /*
   * Commands for setting and displaying the registered port name.
   */
  add_cmd ("port-name", class_support, port_name_set_command,
	   "Set the registered name of GDB's message port.", &setlist);
  add_cmd ("port-name", class_support, port_name_show_command,
	   "Show the registered name of GDB's message port.", &showlist);

  /*
   * Exception commands.
   */
  add_info ("exceptions", exception_info,
	    "What debugger does when program gets various exceptions.\n\
Specify an exception number as argument to print info on that\n\
exception only.");

  add_com ("exception", class_run, exception_command,
	   "Specify how to handle an exception.\n\
Args are exception number followed by \"forward\" or \"keep\".\n\
`Forward' means forward the exception to the program's normal exception\n\
handler.\n\
`Keep' means reenter debugger if this exception happens, and GDB maps\n\
the exception to some signal (see info exception)\n\
Normally \"keep\" is used to return to GDB on exception.");
}

void
_initialize_mach3 ()
{
  kern_return_t kr;

  kr = mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_PORT_SET,
			  &gdb_port_set);
  if (kr != KERN_SUCCESS)
    {
      fprintf (stderr, "Unable to allocate GDB's port set: %s.\n",
	       mach_error_string (kr));
      abort ();
    }

  kr = netname_look_up(name_server_port, "", "MachID", &mid_server);
  if (kr != KERN_SUCCESS)
    {
      mid_server = MACH_PORT_NULL;
      
      fprintf (stderr, "Unable to find MachId server: %s.\n",
	       mach_error_string (kr));
      fprintf (stderr, "Some features disabled...\n");
    }
  
  mid_auth = mach_privileged_host_port();
  if (mid_auth == MACH_PORT_NULL)
    mid_auth = mach_task_self();
  
  obstack_init (port_chain_obstack);

  /* Allocate message port */
  kr = mach_port_allocate (mach_task_self (), MACH_PORT_RIGHT_RECEIVE,
			   &gdb_message_port);
  if (kr != KERN_SUCCESS)
    {
      fprintf (stderr, "Unable to create message port: %s.\n",
	       mach_error_string (kr));
      abort ();
    }
  else
    {
      char buf[ MAX_NAME_LEN ];

      kr = mach_port_move_member(mach_task_self (), gdb_message_port,
				 gdb_port_set);
      if (kr != KERN_SUCCESS)
        {
	  fprintf (stderr, "Unable to add message port to port set: %s.\n",
		   mach_error_string (kr));
	  abort ();
	}

      kr = mach_port_insert_right (mach_task_self (), gdb_message_port,
				   gdb_message_port, MACH_MSG_TYPE_MAKE_SEND);
      if (kr != KERN_SUCCESS)
	{
	  fprintf (stderr, "Unable to add send right to message port: %s.\n",
		   mach_error_string (kr));
	  abort ();
	}

      /*
       * No way to change message port name currently.
       */
      sprintf (buf, "%d", getpid ());
      port_name_set_command (buf, FALSE);
    }
  
  /*
   * Set for handling U*x signals.
   */
  gdb_main_thread_id = cthread_self ();
  cthread_detach(cthread_fork(mach3_thread, NULL));

  signal (SIGCHLD, sigchld_handler);

  add_mach_specific_commands ();

  add_target (&mach3_ops);
}
